# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.plugins.common import constants as p_constants
from neutron.plugins.ml2 import driver_api as api
from oslo_log import log

from networking_powervm.plugins.ibm.agent.powervm import constants as pconst
from networking_powervm.plugins.ml2.drivers import mech_pvm_base

LOG = log.getLogger(__name__)


class PvmSEAMechanismDriver(mech_pvm_base.PvmMechanismDriverBase):
    """Attach to networks using PowerVM Shared Ethernet agent.

    The PvmSEAMechanismDriver integrates the ml2 plugin with the
    PowerVM Shared Ethernet Agent.
    """

    def __init__(self):
        super(PvmSEAMechanismDriver, self).__init__(pconst.AGENT_TYPE_PVM_SEA,
                                                    pconst.VIF_TYPE_PVM_SEA)

    def try_to_bind_segment_for_agent(self, context, segment, agent):
        # When this method is called, the parent should ideally be calling
        # down to the agent to state that the port was updated.  However,
        # it appears this isn't flowing properly.  This makes sure the
        # port is passed down to the agent.
        bindable = (super(PvmSEAMechanismDriver, self).
                    try_to_bind_segment_for_agent(context, segment, agent))
        if bindable:
            self.rpc_publisher.port_update(context._plugin_context,
                                           context._port,
                                           segment[api.NETWORK_TYPE],
                                           segment[api.SEGMENTATION_ID],
                                           segment[api.PHYSICAL_NETWORK])
        return bindable

    def get_allowed_network_types(self, agent=None):
        """Get a list of allowed network types.

        The SEA agent supports only the VLAN network type.

        :param agent: Not used
        :return: List of neutron.plugins.common.constants.TYPE_* strings.
        """
        return [p_constants.TYPE_VLAN]



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_powervm.plugins.ml2.drivers import mech_pvm_sea
from networking_powervm.plugins.ml2.drivers import mech_pvm_sriov
from networking_powervm.tests.unit.plugins.ibm.powervm import base

from neutron.plugins.ml2 import driver_api as api


class BaseTestPvmMechDriver(base.BasePVMTestCase):

    def setUp(self):
        super(BaseTestPvmMechDriver, self).setUp()
        rpc_p = mock.patch('neutron.plugins.ml2.rpc.AgentNotifierApi')
        self.mock_rpc = rpc_p.start()
        self.addCleanup(rpc_p.stop)

        self.context = mock.Mock(current={})
        self.segment = {api.ID: 'id', api.SEGMENTATION_ID: 'seg_id'}
        self.agent = {'host': 'agent_host',
                      'configurations': {
                          'bridge_mappings': {
                              'the_network': ['p1', 'p2'],
                              'the_other_network': ['p3', 'p4']},
                          'default_redundancy': '4',
                          'default_capacity': 'None'}}

    def verify_check_segment_for_agent(self, supp_net_types):
        """Validates that the VLAN type is supported by the agent.

        Covers get_allowed_network_types and physnet_in_mappings.

        :param supp_net_types: Set of network type strings (see
                               neutron.plugins.common.constants.TYPE_*)
                               supported by the mech driver.
        """
        all_net_types = {'flat', 'vlan', 'geneve', 'gre', 'local', 'vxlan',
                         'none'}
        unsupp_net_types = all_net_types - supp_net_types
        seg = self.segment
        agt = self.agent
        for net_type in supp_net_types:
            # Fail for valid net type when network isn't in bridge mappings.
            seg[api.PHYSICAL_NETWORK] = 'bogus_network'
            seg[api.NETWORK_TYPE] = net_type
            self.assertFalse(self.mech_drv.check_segment_for_agent(seg, agt))
            # Succeed when the network is in the bridge mappings
            seg[api.PHYSICAL_NETWORK] = 'the_other_network'
            self.assertTrue(self.mech_drv.check_segment_for_agent(seg, agt))

        # Bogus network types fail even with a network in bridge mappings
        seg[api.PHYSICAL_NETWORK] = 'the_network'
        for net_type in unsupp_net_types:
            seg[api.NETWORK_TYPE] = net_type
            self.assertFalse(self.mech_drv.check_segment_for_agent(seg, agt))

    @mock.patch('neutron.plugins.ml2.drivers.mech_agent.'
                'SimpleAgentMechanismDriverBase.check_segment_for_agent')
    def bad_bind_segment_for_agent(self, mock_cs4a):
        # Binding fails when check_segment_for_agent fails
        mock_cs4a.return_value = False
        self.assertFalse(self.mech_drv.try_to_bind_segment_for_agent(
            'context', self.segment, 'agent'))
        mock_cs4a.assert_called_once_with(self.segment, 'agent')

    @mock.patch('neutron.plugins.ml2.drivers.mech_agent.'
                'SimpleAgentMechanismDriverBase.check_segment_for_agent')
    def good_bind_segment_for_agent(self, mock_cvd, mock_cs4a):
        # Binding succeeds, and calls set_binding and customize_vif_details,
        # when check_segment_for_agent succeeds
        mock_cs4a.return_value = True
        self.assertTrue(self.mech_drv.try_to_bind_segment_for_agent(
            self.context, self.segment, 'agent'))
        self.context.set_binding.assert_called_once_with(
            'id', self.mech_drv.vif_type, mock_cvd.return_value)
        mock_cvd.assert_called_once_with(self.context, self.segment, 'agent')

    def verify_vif_details(self):
        self.segment[api.SEGMENTATION_ID] = 'the_seg_id'
        ret = self.mech_drv.customize_vif_details(
            self.context, self.segment, self.agent)
        self.assertFalse(ret['port_filter'])
        self.assertEqual('the_seg_id', ret['vlan'])
        # With no segmentation ID, vlan is None
        del self.segment[api.SEGMENTATION_ID]
        ret = self.mech_drv.customize_vif_details(
            self.context, self.segment, self.agent)
        self.assertFalse(ret['port_filter'])
        self.assertIsNone(ret['vlan'])
        return ret


class TestPvmSeaMechDriver(BaseTestPvmMechDriver):
    def setUp(self):
        super(TestPvmSeaMechDriver, self).setUp()
        self.mech_drv = mech_pvm_sea.PvmSEAMechanismDriver()

    def test_init(self):
        self.assertEqual('PowerVM Shared Ethernet agent',
                         self.mech_drv.agent_type)
        self.assertEqual(['normal'], self.mech_drv.supported_vnic_types)
        self.assertEqual('pvm_sea', self.mech_drv.vif_type)
        self.assertEqual(self.mock_rpc.return_value,
                         self.mech_drv.rpc_publisher)

    def test_check_segment_for_agent(self):
        self.verify_check_segment_for_agent({'vlan'})

    @mock.patch('networking_powervm.plugins.ml2.drivers.mech_pvm_base.'
                'PvmMechanismDriverBase.customize_vif_details')
    def test_try_to_bind_segment_for_agent(self, mock_cvd):
        # Not bindable - no port_update
        self.bad_bind_segment_for_agent()
        self.mock_rpc.return_value.port_update.assert_not_called()

        # Bindable - port updated
        self.segment[api.NETWORK_TYPE] = 'net_type'
        self.segment[api.PHYSICAL_NETWORK] = 'phys_net'
        self.good_bind_segment_for_agent(mock_cvd)
        self.mock_rpc.return_value.port_update.assert_called_once_with(
            self.context._plugin_context, self.context._port, 'net_type',
            'seg_id', 'phys_net')

    def test_vif_details(self):
        self.verify_vif_details()


class TestPvmSriovMechDriver(BaseTestPvmMechDriver):
    def setUp(self):
        super(TestPvmSriovMechDriver, self).setUp()
        self.mech_drv = mech_pvm_sriov.PvmSRIOVMechanismDriver()

    def test_init(self):
        self.assertEqual('PowerVM SR-IOV Ethernet agent',
                         self.mech_drv.agent_type)
        self.assertEqual(['direct'], self.mech_drv.supported_vnic_types)
        self.assertEqual('pvm_sriov', self.mech_drv.vif_type)
        self.assertEqual(self.mock_rpc.return_value,
                         self.mech_drv.rpc_publisher)

    def test_check_segment_for_agent(self):
        self.verify_check_segment_for_agent({'vlan', 'flat'})

    @mock.patch('networking_powervm.plugins.ml2.drivers.mech_pvm_sriov.'
                'PvmSRIOVMechanismDriver.customize_vif_details')
    def test_try_to_bind_segment_for_agent(self, mock_cvd):
        self.bad_bind_segment_for_agent()
        self.good_bind_segment_for_agent(mock_cvd)

    def test_vif_details(self):
        # No profile in context, mismatched network in segment
        self.segment['physical_network'] = 'bogus_net'
        vif_dets = self.verify_vif_details()
        self.assertEqual('bogus_net', vif_dets['physical_network'])
        # No ports
        self.assertEqual([], vif_dets['physical_ports'])
        # default_redundancy from agent config
        self.assertEqual(4, vif_dets['redundancy'])
        # platform default capacity
        self.assertIsNone(vif_dets['capacity'])

        # Ensure non-None capacity default comes through
        self.agent['configurations']['default_capacity'] = '0.04'
        vif_dets = self.verify_vif_details()
        self.assertEqual(0.04, vif_dets['capacity'])
        self.assertEqual('bogus_net', vif_dets['physical_network'])

        # Now with proper values set in profile & segment
        self.segment['physical_network'] = 'the_other_network'
        self.context.current = {'binding:profile': {'vnic_required_vfs': "10",
                                                    'capacity': "0.16"}}
        vif_dets = self.verify_vif_details()
        self.assertEqual(['p3', 'p4'], vif_dets['physical_ports'])
        self.assertEqual(10, vif_dets['redundancy'])
        self.assertEqual(0.16, vif_dets['capacity'])
        self.assertEqual('the_other_network', vif_dets['physical_network'])



# Copyright 2014, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
import sys

eventlet.monkey_patch()

from oslo_config import cfg
from oslo_log import log as logging

from neutron.common import config as n_config
from neutron.conf.agent import common as a_config
from pypowervm.tasks import network_bridger as net_br
from pypowervm.wrappers import logical_partition as pvm_lpar
from pypowervm.wrappers import network as pvm_net

from networking_powervm._i18n import _LI
from networking_powervm._i18n import _LW
from networking_powervm.plugins.ibm.agent.powervm import agent_base
from networking_powervm.plugins.ibm.agent.powervm import constants as p_const
from networking_powervm.plugins.ibm.agent.powervm import prov_req as preq
from networking_powervm.plugins.ibm.agent.powervm import utils


LOG = logging.getLogger(__name__)


agent_opts = [
    cfg.StrOpt('bridge_mappings',
               default='',
               help='The Network Bridge mappings (defined by the SEA) that '
                    'describe how the neutron physical networks map to the '
                    'Shared Ethernet Adapters.'
                    'Format: <ph_net1>:<sea1>:<vio1>,<ph_net2>:<sea2>:<vio2> '
                    'Example: default:ent5:vios_1,speedy:ent6:vios_1'),
    cfg.BoolOpt('automated_powervm_vlan_cleanup', default=True,
                help='Determines whether or not the VLANs will be removed '
                     'from the Network Bridge if a VM is removed and it is '
                     'the last VM on the system to use that VLAN.  By '
                     'default, will clean up VLANs to improve the overall '
                     'system performance (by reducing broadcast domain).  '
                     'Will only apply to VLANs not on the primary PowerVM '
                     'virtual Ethernet adapter of the SEA.')
]


cfg.CONF.register_opts(agent_opts, "AGENT")
a_config.register_agent_state_opts_helper(cfg.CONF)
a_config.register_root_helper(cfg.CONF)

ACONF = cfg.CONF.AGENT

VIF_TYPE_PVM_SEA = 'pvm_sea'


class SharedEthernetNeutronAgent(agent_base.BasePVMNeutronAgent):
    """
    Provides VLAN networks for the PowerVM platform that run accross the
    Shared Ethernet within the Virtual I/O Servers.  Designed to be compatible
    with the ML2 Neutron Plugin.
    """

    @property
    def agent_id(self):
        return 'sea-agent-%s' % cfg.CONF.host

    @property
    def agent_binary_name(self):
        """Name of the executable under which the SEA agent runs."""
        return p_const.AGENT_BIN_SEA

    @property
    def agent_type(self):
        return p_const.AGENT_TYPE_PVM_SEA

    @property
    def vif_wrapper_class(self):
        return pvm_net.CNA

    @property
    def vif_type(self):
        return VIF_TYPE_PVM_SEA

    def parse_bridge_mappings(self):
        return utils.parse_sea_mappings(self.adapter, self.host_uuid,
                                        ACONF.bridge_mappings)

    def heal_and_optimize(self):
        """Heals the system's network bridges and optimizes.

        Will query neutron for all the ports in use on this host.  Ensures that
        all of the VLANs needed for those ports are available on the correct
        network bridge.

        Finally, it optimizes the system by removing any VLANs that may no
        longer be required.  The VLANs that are removed must meet the following
        conditions:
         - Are not in use by ANY virtual machines on the system.  OpenStack
           managed or not.
         - Are not part of the primary load group on the Network Bridge.
        """
        LOG.info(_LI("Running the heal and optimize flow."))

        # Get a map of all the partitions and their CNAs.
        all_lpar_cnas = utils.list_vifs(self.adapter, self.vif_wrapper_class,
                                        include_vios_and_mgmt=True)
        # There can be CNAs (non-trunk) on VIOSes as well.  They should be
        # taken into account for the overall vifs.
        overall_vifs = [vif for vifs in all_lpar_cnas.values() for vif in vifs]
        # The lpar_cna_map includes only the client VMs.  Don't include mgmt.
        lpar_cna_map = {par_w: vifs for par_w, vifs in all_lpar_cnas.items()
                        if isinstance(par_w, pvm_lpar.LPAR) and
                        not par_w.is_mgmt_partition}

        # Build out all of the devices that we have available to us.  Some
        # may be on the system but not part of OpenStack.  Those get ignored.
        prov_reqs = preq.ProvisionRequest.for_wrappers(self, lpar_cna_map,
                                                       preq.PLUG)
        # Dictionary of the required VLANs on the Network Bridge
        nb_req_vlans = {}
        nb_wraps = utils.list_bridges(self.adapter, self.host_uuid)
        for nb_wrap in nb_wraps:
            nb_req_vlans[nb_wrap.uuid] = set()

        # Call down to the provision.  This will call device up on the
        # requests.
        self.provision_devices(prov_reqs)

        # Make sure that the provision requests VLAN is captured in the
        # nb_req_vlans list...so that the VLAN is not accidentally removed.
        for req in prov_reqs:
            nb_uuid, req_vlan = self._get_nb_and_vlan(req.rpc_device,
                                                      emit_warnings=False)
            nb_req_vlans[nb_uuid].add(req_vlan)

        # We should clean up old VLANs as well.  However, we only want to clean
        # up old VLANs that are not in use by ANYTHING in the system.
        #
        # The first step is to identify the VLANs that are needed.  That can
        # be done by extending our nb_req_vlans map.
        #
        # We first extend that map by listing all the VMs on the system
        # (whether managed by OpenStack or not) and then seeing what Network
        # Bridge uses them.
        vswitch_map = utils.get_vswitch_map(self.adapter, self.host_uuid)
        for client_adpt in overall_vifs:
            nb = utils.find_nb_for_cna(nb_wraps, client_adpt, vswitch_map)
            # Could occur if a system is internal only.
            if nb is None:
                LOG.debug("Client Adapter with mac %s is internal only.",
                          client_adpt.mac)
                continue

            # Make sure that it is on the nb_req_vlans list, as it is now
            # considered required.
            nb_req_vlans[nb.uuid].add(client_adpt.pvid)

            # Extend for each additional vlans as well
            for addl_vlan in client_adpt.tagged_vlans:
                nb_req_vlans[nb.uuid].add(addl_vlan)

        # The list of required VLANs on each network bridge also includes
        # everything on the primary VEA.
        for nb in nb_wraps:
            prim_ld_grp = nb.load_grps[0]
            vlans = [prim_ld_grp.pvid]
            vlans.extend(prim_ld_grp.tagged_vlans)
            for vlan in vlans:
                nb_req_vlans[nb.uuid].add(vlan)

        # If the configuration is set.
        if ACONF.automated_powervm_vlan_cleanup:
            self._cleanup_unused_vlans(nb_wraps, nb_req_vlans)

    def _cleanup_unused_vlans(self, nb_wraps, nb_req_vlans):
        cur_delete = 0

        # Loop through and remove VLANs that are no longer needed.
        for nb in nb_wraps:
            # Join the required vlans on the network bridge (already in
            # use) with the pending VLANs.
            req_vlans = nb_req_vlans[nb.uuid]

            # Get ALL the VLANs on the bridge
            existing_vlans = set(nb.list_vlans())

            # To determine the ones no longer needed, subtract from all the
            # VLANs the ones that are no longer needed.
            vlans_to_del = existing_vlans - req_vlans
            for vlan_to_del in vlans_to_del:
                if cur_delete < 3:
                    LOG.warning(_LW("Cleaning up VLAN %(vlan)s from the "
                                    "system. It is no longer in use."),
                                {'vlan': vlan_to_del})
                    net_br.remove_vlan_from_nb(self.adapter, self.host_uuid,
                                               nb.uuid, vlan_to_del)
                else:
                    # We don't want to block on optimization for too long.
                    # Each VLAN clean up can take ~2 seconds, so if we do
                    # three of them, then that blocks deploys for about 6
                    # seconds.  We generally don't clean out VLANs that often
                    # but just in case, we get a rush of them, this ensures
                    # we don't block provision requests that are actually going
                    # on in the system.
                    LOG.warning(_LW(
                        "System identified that VLAN %(vlan)s is unused. "
                        "However three VLAN clean ups have already occurred "
                        "in this pass. Will clean up in next optimization "
                        "pass."), {'vlan': vlan_to_del})
                cur_delete += 1

    def provision_devices(self, requests):
        """Will ensure that the VLANs are on the NBs for the edge devices.

        Takes in a set of ProvisionRequests.  From those devices, determines
        the correct network bridges and their appropriate VLANs.  Then calls
        down to the pypowervm API to ensure that the required VLANs are
        on the appropriate ports.

        Will also ensure that the client side adapter is updated with the
        correct VLAN.

        :param requests: A list of ProvisionRequest objects.
        """
        # Only handle 'plug' requests.
        plug_reqs = {req for req in requests if req.action == preq.PLUG}
        nb_to_vlan = {}
        for p_req in plug_reqs:
            # Break the ports into their respective lists broken down by
            # Network Bridge.
            nb_uuid, vlan = self._get_nb_and_vlan(p_req.rpc_device,
                                                  emit_warnings=True)

            # A warning message will be printed to user if this were to occur
            if nb_uuid is None:
                continue

            if nb_to_vlan.get(nb_uuid) is None:
                nb_to_vlan[nb_uuid] = set()

            nb_to_vlan[nb_uuid].add(vlan)

        # For each bridge, make sure the VLANs are serviced.
        for nb_uuid in nb_to_vlan:
            net_br.ensure_vlans_on_nb(self.adapter, self.host_uuid, nb_uuid,
                                      nb_to_vlan.get(nb_uuid))

        # Now that the bridging is complete, let the superclass mark them as up
        super(SharedEthernetNeutronAgent, self).provision_devices(plug_reqs)
        LOG.debug('Successfully provisioned SEA VLANs for new devices.')

    def _get_nb_and_vlan(self, dev, emit_warnings=False):
        """Parses bridge mappings to find a match for the device passed in.
        :param dev: Neutron device to find a match for
        :param emit_warnings: (Optional) Defaults to False.  If true, will emit
                              a warning if the configuration is off.
        :return: UUID of the NetBridge
        :return: vlan for the neutron device
        """
        nb_uuid = self.br_map.get(dev.get('physical_network'))
        if not nb_uuid and emit_warnings:
            LOG.warning(_LW("Unable to determine the Network Bridge (Shared "
                            "Ethernet Adapter) for physical network %s.  Will "
                            "be unable to determine appropriate provisioning "
                            "action."), dev.get('physical_network'))
        return nb_uuid, dev.get('segmentation_id')


def main():
    # Read in the command line args
    n_config.init(sys.argv[1:])
    n_config.setup_logging()

    # Build then run the agent
    agent = SharedEthernetNeutronAgent()
    LOG.info(_LI("Shared Ethernet Agent initialized and running"))
    agent.rpc_loop()


if __name__ == "__main__":
    main()



# Copyright 2015, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import eventlet
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall

from neutron.agent import rpc as agent_rpc
from neutron.common import topics
from neutron.conf.agent import common as a_config
from neutron_lib import constants as q_const
from neutron_lib import context as ctx
from pypowervm import adapter as pvm_adpt
from pypowervm.helpers import log_helper as log_hlp
from pypowervm.helpers import vios_busy as vio_hlp
from pypowervm.wrappers import event as pvm_evt
from pypowervm.wrappers import managed_system as pvm_ms

from networking_powervm._i18n import _
from networking_powervm._i18n import _LE
from networking_powervm._i18n import _LI
from networking_powervm._i18n import _LW
from networking_powervm.plugins.ibm.agent.powervm import prov_req as preq
from networking_powervm.plugins.ibm.agent.powervm import utils

eventlet.monkey_patch()


LOG = logging.getLogger(__name__)

agent_opts = [
    cfg.IntOpt('exception_interval', default=5,
               help=_("The number of seconds agent will wait between "
                      "polling when exception is caught")),
    cfg.IntOpt('heal_and_optimize_interval', default=1800,
               help=_('The number of seconds the agent should wait between '
                      'heal/optimize intervals.')),
]

cfg.CONF.register_opts(agent_opts, "AGENT")
a_config.register_agent_state_opts_helper(cfg.CONF)
a_config.register_root_helper(cfg.CONF)

ACONF = cfg.CONF.AGENT


# Event types requiring refetch of all VIFs for all LPARs.
FULL_REFETCH_EVENTS = (
    pvm_evt.EventType.CACHE_CLEARED, pvm_evt.EventType.MISSING_EVENTS,
    pvm_evt.EventType.NEW_CLIENT)
# Event types affecting a single object. The object's URI is in Event.data.
SINGLE_OBJ_EVENTS = (
    pvm_evt.EventType.INVALID_URI, pvm_evt.EventType.ADD_URI,
    pvm_evt.EventType.MODIFY_URI, pvm_evt.EventType.HIDDEN_URI,
    pvm_evt.EventType.VISIBLE_URI, pvm_evt.EventType.CUSTOM_CLIENT_EVENT,
    pvm_evt.EventType.DELETE_URI)


class PVMPluginApi(agent_rpc.PluginApi):
    pass


class VIFEventHandler(pvm_adpt.WrapperEventHandler):
    """Listens for Events from the PowerVM API that could be network events.

    This event handler will be invoked by the PowerVM API when something occurs
    on the system.  This event handler will determine if it could have been
    related to a network change.  If so, then it will add a ProvisionRequest
    to the processing queue.
    """

    def __init__(self, agent):
        self.agent = agent
        self.adapter = self.agent.adapter
        self.just_started = True

    def _refetch_all(self, prov_req_set):
        """Populate prov_req_set with ProvisionRequests for all LPARs' VIFs.

        :param prov_req_set: Set of ProvisionRequests, updated by this method.
        """
        # Create 'plug' requests for all LPARs' VIFs.  (No VIOSes or management
        # partition.)
        lpar_vifs = utils.list_vifs(self.adapter, self.agent.vif_wrapper_class)
        prov_reqs = preq.ProvisionRequest.for_wrappers(self.agent, lpar_vifs,
                                                       preq.PLUG)
        # Obliterate any 'plug' requests; they'll be replaced below if they
        # still exist.  Leave 'unplug' requests, which should represent VIFs
        # that will be absent from the comprehensive refetch result.
        rms = {req for req in prov_req_set if req.action == preq.PLUG}
        LOG.debug("Removing all existing plug requests: %s", rms)
        prov_req_set -= rms
        # Add in the wrapper-based 'plug' requests generated above.
        # They'll be ignored later if there's no matching neutron port.
        LOG.debug("Adding new wrapper-based plug requests: %s",
                  [str(prov_req) for prov_req in prov_reqs])
        prov_req_set |= set(prov_reqs)

    def _process_event(self, event, prov_req_set):
        """Process a PowerVM event, folding into prov_req_set as appropriate.

        :param event: The pypowervm.wrappers.event.Event to be processed.
        :return: True if the event resulted in an actionable ProvisionRequest
                 being added; False otherwise.
        """
        prov_req = preq.ProvisionRequest.for_event(self.agent, event)
        if prov_req is None:
            # Nothing to do
            return False

        # Consolidate requests.  Collapse if same action.  Replace if opposite
        # action.  ProvisionRequest's __eq__ will hit if the MAC and LPAR UUID
        # match; so this add will find hits for either action.
        rms = {prq for prq in prov_req_set if prq == prov_req}
        LOG.debug("Consolidating - removing requests: %s",
                  [str(rm_preq) for rm_preq in rms])
        prov_req_set -= rms
        LOG.debug("Adding new event-based request: %s", str(prov_req))
        prov_req_set.add(prov_req)

    def process(self, events):
        """Process a list of REST events.

        :param events: A list of pypowervm.wrappers.event.Event wrappers to
                       handle.
        """
        prov_req_set = set()
        do_heal = False
        for event in events:
            if event.etype in FULL_REFETCH_EVENTS:
                # On initial startup, we'll get a NEW_CLIENT event, *but*
                # heal_and_optimize will also do a full sweep; so skip this
                # that first time.
                if not self.just_started:
                    # Full refetch of all LPARs' VIFs.  Subsequent events in
                    # this iteration may still add/remove entries.
                    self._refetch_all(prov_req_set)
            elif event.etype in SINGLE_OBJ_EVENTS:
                self._process_event(event, prov_req_set)

            if self.agent.is_hao_event(event):
                LOG.info(_LI("Received heal-and-optimize event: %s"),
                         str(event))
                do_heal = True

            self.just_started = False

        # Do heal_and_optimize if requested.
        if do_heal:
            self.agent.heal_and_optimize()

        # Process any port requests accumulated above.
        self.agent.provision_devices(prov_req_set)


class BasePVMNeutronAgent(object):
    """Baseline PowerVM Neutron Agent class for extension.

    The ML2 agents have a common RPC polling framework and API callback
    mechanism.  This class provides the baseline so that other children
    classes can extend and focus on their specific functions rather than
    integration with the RPC server.
    """

    # This agent supports RPC Version 1.0.  Though agents don't boot unless
    # 1.1 or higher is specified now.
    # For reference:
    #  1.0 Initial version
    #  1.1 Support Security Group RPC
    #  1.2 Support DVR (Distributed Virtual Router) RPC
    RPC_API_VERSION = '1.1'

    @abc.abstractproperty
    def agent_id(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def agent_binary_name(self):
        """Name of the executable under which the (subclass) agent runs."""
        raise NotImplementedError()

    @abc.abstractproperty
    def agent_type(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def vif_type(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def vif_wrapper_class(self):
        """The pypowervm wrapper class for the VIF-ish type the agent handles.

        E.g. pypowervm.wrappers.network.CNA, pypowervm.wrappers.iocard.VNIC.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def parse_bridge_mappings(self):
        """This method should return the bridge mappings dictionary.

        The pypowervm adapter will be initialized before this method is called.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def heal_and_optimize(self):
        """Ensures that the bridging supports all the needed ports.

        This method is invoked periodically (not on every RPC loop).  Its
        purpose is to ensure that the bridging supports every client VM
        properly.  If possible, it should also optimize the connections.
        """
        raise NotImplementedError()

    def is_hao_event(self, evt):
        """Determines if an Event warrants a heal_and_optimize.

        :param evt: A pypowervm.wrappers.event.Event wrapper to inspect.
        :return: True if heal_and_optimize should be invoked as a result of
                 this event; False otherwise.
        """
        return False

    def customize_agent_state(self):
        """Perform subclass-specific adjustments to self.agent_state."""
        pass

    def setup_adapter(self):
        """Configure the pypowervm adapter.

        This method assigns a valid pypowervm.adapter.Adapter to the
        self.adapter instance variable.
        """
        # Attempt multiple times in case the REST server is starting.
        self.adapter = pvm_adpt.Adapter(
            pvm_adpt.Session(conn_tries=300),
            helpers=[log_hlp.log_helper, vio_hlp.vios_busy_retry_helper])

    def __init__(self):
        """Create the PVM neutron agent. """
        self.adapter = None
        self.setup_adapter()
        self.msys = pvm_ms.System.get(self.adapter)[0]
        self.host_uuid = self.msys.uuid

        # Get the bridge mappings
        self.br_map = self.parse_bridge_mappings()

        self.agent_state = {
            'binary': self.agent_binary_name,
            'host': cfg.CONF.host,
            'topic': q_const.L2_AGENT_TOPIC,
            'configurations': {'bridge_mappings': self.br_map},
            'agent_type': self.agent_type,
            'start_flag': True}
        self.customize_agent_state()

        # Set Up RPC to Server
        self._setup_rpc()

        # Add VIF event handler to the session.
        evt_listener = self.adapter.session.get_event_listener()
        self._vif_event_handler = VIFEventHandler(self)
        evt_listener.subscribe(self._vif_event_handler)

    def port_update(self, context, **kwargs):
        """RPC callback indicating that a port has been updated within Neutron.

        This callback is registered as part of _setup_rpc.  It is invoked by
        the controller when a port is updated in Neutron.

        This is currently a no-op.
        """
        port = kwargs['port']
        LOG.debug('Neutron API port_update RPC received for port '
                  'id=%(port_id)s, mac=%(mac)s, instance=%(inst_id)s.',
                  {'port_id': port['id'], 'mac': port.get('mac_address'),
                   'inst_id': port['device_id']})

    def _setup_rpc(self):
        """Registers the RPC consumers for the plugin."""
        self.topic = topics.AGENT
        self.plugin_rpc = PVMPluginApi(topics.PLUGIN)
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.PLUGIN)

        self.context = ctx.get_admin_context_without_session()

        # Define the listening consumers for the agent.  ML2 only supports
        # these two update types.
        consumers = [[topics.PORT, topics.UPDATE]]

        self.connection = agent_rpc.create_consumers([self],
                                                     self.topic,
                                                     consumers)

        # Report interval is for the agent health check.
        report_interval = ACONF.report_interval
        if report_interval:
            hb = loopingcall.FixedIntervalLoopingCall(self._report_state)
            hb.start(interval=report_interval)

    def _report_state(self):
        """Reports the state of the agent back to the controller.

        Controller knows that if a response isn't provided in a certain period
        of time then the agent is dead.  This call simply tells the controller
        that the agent is alive.
        """
        # TODO(thorst) provide some level of devices connected to this agent.
        try:
            device_count = 0
            self.agent_state.get('configurations')['devices'] = device_count
            LOG.debug("Reporting agent state to neutron: %s", self.agent_state)
            self.state_rpc.report_state(self.context,
                                        self.agent_state)
            self.agent_state.pop('start_flag', None)
        except Exception:
            LOG.exception(_LE("Failed reporting state!"))

    def update_device_up(self, device):
        """Calls back to neutron that a device is alive.

        :param device: The device detail from get_device[s]_details[_list].
        """
        LOG.info(_LI("Sending device up to Neutron for %(dev)s"),
                 {'dev': device['device']})
        self.plugin_rpc.update_device_up(self.context, device['device'],
                                         self.agent_id, cfg.CONF.host)

    def update_device_down(self, device):
        """Calls back to neutron that a device is down.

        :param device: The device detail from get_device[s]_details[_list].
        """
        LOG.warning(_LW("Sending device DOWN to Neutron for %(dev)s"),
                    {'dev': device['device']})
        self.plugin_rpc.update_device_down(self.context, device['device'],
                                           self.agent_id, cfg.CONF.host)

    def get_device_details(self, device_mac):
        """Returns a neutron device for a given mac address.

        :param device_mac: The neutron mac addresses for the device to get.
        :return: The device from neutron.
        """
        return self.plugin_rpc.get_device_details(
            self.context, utils.norm_mac(device_mac), self.agent_id)

    def get_devices_details_list(self, device_macs):
        """Returns list of neutron devices for a list of mac addresses.

        :param device_macs: List of neutron mac addresses for the devices to
                            get.
        :return: The list of devices from neutron.
        """
        return self.plugin_rpc.get_devices_details_list(
            self.context, [utils.norm_mac(mac) for mac in device_macs],
            self.agent_id)

    def provision_devices(self, requests):
        """Invoked when a set of new Neutron ports has been detected.

        This method should provision the bridging for the new devices.

        The subclass implementation may be non-blocking.  This means, if it
        will take a very long time to provision, or has a dependency on
        another action (ex. client VIF needs to be created), then it should
        run in a separate worker thread.

        Because of the non-blocking nature of the method, it is required that
        the child class updates the device state upon completion of the device
        provisioning.  This can be done with the agent's update_device_up/_down
        methods, or by invoking this default implementation.

        :param requests: A list of ProvisionRequest objects.
        """
        LOG.debug("Provisioning %d devices.", len(requests))
        for p_req in requests:
            if p_req.action == preq.PLUG:
                self.update_device_up(p_req.rpc_device)
            elif p_req.action == preq.UNPLUG:
                self.update_device_down(p_req.rpc_device)
            else:
                LOG.warning(_LW("Ignoring provision request with unknown "
                                "action: %s"), str(p_req))

    def rpc_loop(self):
        """
        Runs a check periodically to determine if new ports were added or
        removed.  Will call down to appropriate methods to determine correct
        course of action.
        """
        while True:
            try:
                # If the loop interval has passed, heal and optimize
                LOG.debug("Performing heal and optimization of system.")
                self.heal_and_optimize()
                time.sleep(ACONF.heal_and_optimize_interval)

            except Exception as e:
                LOG.exception(e)
                LOG.warning(_LW("Error has been encountered and logged.  The "
                                "agent will retry again."))
                # sleep for a while and re-loop
                time.sleep(ACONF.exception_interval)



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""ProvisionRequest class and related artifacts."""
import six
import time

from neutron.conf.agent import common as a_config
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from pypowervm import util as pvm_util
from pypowervm.wrappers import event as pvm_evt

from networking_powervm._i18n import _LI
from networking_powervm.plugins.ibm.agent.powervm import utils

a_config.register_agent_state_opts_helper(cfg.CONF)
a_config.register_root_helper(cfg.CONF)

LOG = logging.getLogger(__name__)

# Time out waiting for a port's VIF to be plugged after 20 minutes.
PORT_TIMEOUT_S = 20 * 60

# Provisioning actions
PLUG = 'plug'
UNPLUG = 'unplug'

# Event provider ID for nova_powervm.virt.powervm.vif
EVENT_PROVIDER_NOVA_PVM_VIF = 'NOVA_PVM_VIF'


class ProvisionRequest(object):
    """A request for a Neutron Port to be provisioned.

    The RPC device details provide some additional details that the port does
    not necessarily have, and vice versa.  This meshes together the required
    aspects into a single element.
    """

    def __init__(self, action, device_detail, lpar_uuid, vif_type=None):
        """Create a ProvisionRequest for a neutron device associated with a VM.

        Consumers should not call this directly, but should instead use one of
        the factory methods: for_wrappers, for_event, or for_port.

        :param action: One of PLUG or UNPLUG, indicating whether this request
                       should result in the device being marked up or down.
        :param device_detail: The neutron device detail dict returned from
                              the agent's get_device[s]_details[_list].  Should
                              be pre-validated (see utils.device_detail_valid).
        :param lpar_uuid: String UUID of the Logical Partition associated with
                          the device, in PowerVM format (see
                          pypowervm.utils.uuid.convert_uuid_to_pvm).
        :param vif_type: Source of event. It could be pvm_sea or pvm_sriov or
                         others
        """
        self.action = action
        self.mac_address = device_detail.get('mac_address')
        self.rpc_device = device_detail
        self.lpar_uuid = lpar_uuid
        self.created_at = time.time()
        self.vif_type = vif_type

    def __eq__(self, other):
        if not isinstance(other, ProvisionRequest):
            return False

        # Really just need to check the lpar_uuid and mac.  The rest should
        # be static and identical.
        return (other.mac_address == self.mac_address and
                other.lpar_uuid == self.lpar_uuid)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        # Mac addresses should not collide.  This should be sufficient for a
        # hash.  The equals will go just a bit further.
        return hash(self.mac_address)

    def __str__(self):
        return ("ProvisionRequest(action=%(action)s, mac=%(mac)s, "
                "lpar_uuid=%(lpar_uuid)s)" % {'action': self.action,
                                              'mac': self.mac_address,
                                              'lpar_uuid': self.lpar_uuid})

    @classmethod
    def for_wrappers(cls, agent, lpar_vif_map, action):
        """Factory method to produce ProvisionRequests from VIF wrappers.

        :param agent: The neutron agent making the request.
        :param lpar_vif_map: Dict of {lpar_uuid: [vif_w, ...]), where lpar_uuid
                             is the UUID of the LPAR owning the VIF; and each
                             vif_w is a pypowervm wrapper of a VIF-ish type
                             (CNA, VNIC, etc.)
        :param action: What kind of provision requests, either PLUG or UNPLUG.
        :return: A list of new ProvisionRequest instances for each "valid"
                 LPAR/VIF tuple in lpar_vif_list.  Here "valid" means the VIF's
                 corresponding device can be found in neutron, and its instance
                 match the LPAR.
        """
        # Dict mapping {mac: device_detail}
        macs = [vif.mac for vifs in lpar_vif_map.values() for vif in vifs]
        device_details = {dev.get('mac_address'): dev for dev in
                          agent.get_devices_details_list(macs)}
        ret = []
        for lpar, viflist in six.iteritems(lpar_vif_map):
            for vif_w in viflist:
                mac = utils.norm_mac(vif_w.mac)
                if mac not in device_details:
                    # A VIF with no corresponding neutron port
                    continue
                detail = device_details[mac]
                if not utils.device_detail_valid(detail, mac):
                    continue
                LOG.info(_LI(
                    "Creating wrapper-based %(action)s ProvisionRequest for "
                    "%(vif_type)s VIF with MAC %(mac)s associated with LPAR "
                    "%(lpar_name)s (%(lpar_uuid)s)."),
                    {'action': action, 'vif_type': vif_w.schema_type,
                     'mac': vif_w.mac, 'lpar_name': lpar.name,
                     'lpar_uuid': lpar.uuid})
                ret.append(cls(action, detail, lpar.uuid))

        return ret

    @classmethod
    def for_event(cls, agent, event):
        """Factory method to produce a ProvisionRequest for an Event.

        :param agent: The neutron agent making the request.
        :param event: pypowervm.wrappers.event.Event to be processed.
        :return: A new ProvisionRequest.  Returns None if the event is not of
                 interest to the agent.  If the event indicates a PLUG, returns
                 None if a corresponding device can't be found in neutron.
        """
        # Today, we're only handling CUSTOM_CLIENT_EVENTS provided by
        # nova-powervm's vif driver.  In the future, if PowerVM provides
        # official events for VIF types (CNA, VNIC, etc.), this method can be
        # converted to use them.
        if event.etype != pvm_evt.EventType.CUSTOM_CLIENT_EVENT:
            return None
        try:
            edetail = jsonutils.loads(event.detail)
        except (ValueError, TypeError):
            # Not a custom event we recognize
            return None
        if edetail.get('provider') != EVENT_PROVIDER_NOVA_PVM_VIF:
            # Not provided by nova-powervm's vif driver
            return None

        # The actions in the event should match our PLUG/UNPLUG consts, but
        # account for mismatched future versions
        action = edetail['action']
        if action not in (PLUG, UNPLUG):
            LOG.debug("Ignoring event due to unhandled 'action' type.  %s",
                      str(event))
            return None

        device_detail = agent.get_device_details(edetail['mac'])
        if not utils.device_detail_valid(device_detail, edetail['mac']):
            # device_detail_valid logged why
            return None

        # The event data is the URI.  For this kind of event, it looks like:
        # .../LogicalPartition/<LPAR_UUID>/VirtualNICDedicated/<vnic_uuid>
        lpar_uuid = pvm_util.get_req_path_uuid(event.data, preserve_case=True,
                                               root=True)
        vif_type = edetail['type']
        if agent.vif_type != vif_type:
            return None

        LOG.info(_LI("Creating event-based %(action)s ProvisionRequest for "
                     "VIF %(uri)s with MAC %(mac)s associated with LPAR "
                     "%(lpar_uuid)s and source %(vif_type)s."),
                 {'action': edetail['action'], 'uri': event.data,
                  'mac': edetail['mac'], 'lpar_uuid': lpar_uuid,
                  'vif_type': vif_type})
        return cls(edetail['action'], device_detail, lpar_uuid, vif_type)



# Copyright 2014, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Provides a set of utilities for API interaction and Neutron."""

import time

from oslo_log import log as logging

from pypowervm import const as pvm_const
from pypowervm import exceptions as pvm_exc
from pypowervm.helpers import log_helper as pvm_log
from pypowervm.tasks import partition as pvm_par
from pypowervm import util as pvm_util
from pypowervm.utils import retry as pvm_retry
from pypowervm.wrappers import logical_partition as pvm_lpar
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import network as pvm_net
from pypowervm.wrappers import virtual_io_server as pvm_vios

from networking_powervm._i18n import _LI
from networking_powervm._i18n import _LW
from networking_powervm.plugins.ibm.agent.powervm import exceptions as np_exc

LOG = logging.getLogger(__name__)


def parse_sea_mappings(adapter, host_uuid, mapping):
    """This method will parse the sea mappings, and return a UUID map.

    The UUID of the NetworkBridges are required for modification of the
    VLANs that are bridged through the system (via the
    SharedEthernetAdapters). However, UUIDs are not user consumable.  This
    method will read in the string from the CONF file and return a mapping
    for the physical networks.

    Input:
     - <ph_network>:<sea>:<vios_name>,<next ph_network>:<sea2>:<vios_name>
     - Example: default:ent5:vios_lpar,speedy:ent6:vios_lpar

    Output:
    {
      'default': <Network Bridge UUID>, 'speedy': <Network Bridge 2 UUID>
    }

    :param adapter: The pypowervm adapter.
    :param host_uuid: The UUID for the host system.
    :param mapping: The mapping string as defined above to parse.
    :return: The output dictionary described above.
    """
    # Read all the network bridges.
    nb_wraps = list_bridges(adapter, host_uuid)

    if len(nb_wraps) == 0:
        raise np_exc.NoNetworkBridges()
    # Did the user specify the mapping?
    if mapping == '':
        return _parse_empty_bridge_mapping(nb_wraps)

    # Need to find a list of all the VIOSes names to hrefs
    vio_wraps = pvm_vios.VIOS.get(adapter, xag=[pvm_const.XAG.VIO_NET])

    # Response dictionary
    resp = {}

    # Parse the strings
    trios = mapping.split(',')
    for trio in trios:
        # Keys
        # 0 - physical network
        # 1 - SEA name
        # 2 - VIO name
        keys = trio.split(':')

        # Find the VIOS wrapper for the name
        vio_w = next(v for v in vio_wraps if v.name == keys[2])

        # For each network bridge, see if it maps to the SEA name/VIOS href
        matching_nb = None
        for nb_wrap in nb_wraps:
            for sea in nb_wrap.seas:
                sea_vio_uuid = pvm_util.get_req_path_uuid(
                    sea.vio_uri, preserve_case=True)
                if sea.dev_name == keys[1] and sea_vio_uuid == vio_w.uuid:
                    # Found the matching SEA.
                    matching_nb = nb_wrap
                    break

        # Assuming we found a matching SEA, add it to the dictionary
        if matching_nb is not None:
            resp[keys[0]] = matching_nb.uuid
        else:
            raise np_exc.DeviceNotFound(dev=keys[1], vios=keys[2],
                                        phys_net=keys[0])

    return resp


def _parse_empty_bridge_mapping(bridges):
    """Will attempt to derive a bridge mapping if not specified.

    This method is invoked if there is no bridge mapping specified.
    If this happens, it will determine if there is a single Network Bridge
    on the system.  If so, it will assert that the default Neutron
    physical network resides on the singular Network Bridge.

    If there are multiple Network Bridges, an exception is raised.

    This does allow systems to not require the bridge mappings, but it
    is not ideal.

    :param bridges: A list of the network bridges returned via the API.
    :return: The bridge mapping, with a single physical network (default).
    :raises MultiBridgeNoMapping: Thrown if there are multiple Network
                                  Bridges on the system.
    """
    if len(bridges) > 1:
        raise np_exc.MultiBridgeNoMapping()

    LOG.warning(_LW('The bridge_mappings for the agent was not specified.  '
                    'There was exactly one Network Bridge on the system.  '
                    'Agent is assuming the default network is backed by the '
                    'single Network Bridge.'))
    return {'default': bridges[0].uuid}


def norm_mac(mac):
    """
    Will return a MAC Address that normalizes from the pypowervm format
    to the neutron format.

    That means that the format will be converted to lower case and will
    have colons added.

    :param mac: A pypowervm mac address.  Ex. 1234567890AB
    :returns: A mac that matches the standard neutron format.
              Ex. 12:34:56:78:90:ab
    """
    mac = mac.lower().replace(':', '')
    return ':'.join(mac[i:i + 2] for i in range(0, len(mac), 2))


def find_nb_for_cna(nb_wraps, client_adpt, vswitch_map):
    """
    Determines the NetworkBridge (if any) that is supporting a client
    adapter.

    :param nb_wraps: The network bridge wrappers on the system.
    :param client_adpt: The client adapter wrapper.
    :param vswitch_map: Maps the vSwitch IDs to URIs.
                        See 'get_vswitch_map'
    :return The Network Bridge wrapper that is hosting the client adapter.
            If there is not one, None is returned.
    """
    for nb_wrap in nb_wraps:
        # If the vSwitch ID doesn't match the vSwitch on the CNA...don't
        # process
        if vswitch_map.get(nb_wrap.vswitch_id) != client_adpt.vswitch_uri:
            continue

        # If the VLAN is not on the network bridge, then do not process.
        if not nb_wrap.supports_vlan(client_adpt.pvid):
            continue

        # At this point, the client adapter is supported by this network
        # bridge
        return nb_wrap

    # No valid network bridge
    return None


@pvm_retry.retry()
def get_vswitch_map(adapter, host_uuid):
    """Returns a dictionary of vSwitch IDs to their URIs.

    Ex. {'0': 'https://.../VirtualSwitch/<UUID>'}

    :param adapter: The pypowervm adapter.
    :param host_uuid: The UUID for the host system.
    """
    vswitches = pvm_net.VSwitch.get(
        adapter, parent_type=pvm_ms.System, parent_uuid=host_uuid)
    resp = {}
    for vswitch in vswitches:
        resp[vswitch.switch_id] = vswitch.related_href
    return resp


def list_vifs(adapter, vif_class, include_vios_and_mgmt=False):
    """Map of partition:[VIFs] for a specific VIF type (CNA, VNIC, etc.).

    VIOS trunk adapters are never included (even if include_vios_and_mgmt=True)

    :param adapter: The pypowervm adapter.
    :param vif_class: The pypowervm wrapper class for the VIF-ish type to be
                      retrieved (CNA, VNIC, etc.).
    :param include_vios_and_mgmt: If True, the return includes VIFs belonging
                                  to the management partition; AND non-trunk
                                  VIFs belonging to VIOS partitions.  If False,
                                  both of these types are excluded.
    :return: A map of {lpar_w: [vif_w, ...]} where each vif_w is a wrapper
             of the specified vif_class.  The vif_w list may be empty for a
             given lpar_w.
    """
    # Get the VMs to query for.
    LOG.info(_LI("Gathering Virtual Machine wrappers for a list_vifs call. "
                 "Include VIOS and management: %s"), include_vios_and_mgmt)

    # Loop through the VMs
    total_vifs = {}
    for vm_wrap in pvm_par.get_partitions(adapter, lpars=True,
                                          vioses=include_vios_and_mgmt,
                                          mgmt=include_vios_and_mgmt):
        total_vifs[vm_wrap] = _find_vifs(adapter, vif_class, vm_wrap)

    return total_vifs


@pvm_retry.retry(tries=200, delay_func=lambda *a, **k: time.sleep(5),
                 test_func=lambda *a, **k: True)
def _find_vifs(adapter, vif_class, vm_wrap):
    """Return the list of virtual network devices (VIFs) for a partition.

    When vif_class is CNA, this method returns the list of client network
    adapters.  But it defines what a client network adapter is different from
    the pypowervm API.  It does this because pypowervm has an odd definition.

    To the SEA agent, a Client Network Adapter is:
     - ANY pypowervm CNA that is on a LPAR (OpenStack Managed/Unmanaged or
       possibly even the mgmt LPAR)
     - All CNAs on the VIOS partition types that are NOT trunk adapters.

    This method returns those two types because it is often used with the
    heal_and_optimize workflow.  That workflow wants to know all of the VLANs
    that are in use on the system, so that it can trim out any extra VLANs
    from the VIOS type VMs trunk adapters (to reduce the broadcast domain).

    :param adapter: The pypowervm API adapter.
    :param vif_class: The pypowervm wrapper class for the VIF-ish type to be
                      retrieved (CNA, VNIC, etc.).
    :param vm_wrap: The partition wrapper (LPAR or VIOS type) whose VIFs are to
                    be retrieved.
    """
    try:
        vif_list = vif_class.get(
            adapter, parent=vm_wrap, helpers=_remove_log_helper(adapter))

        # This method returns all of the VIF wrappers.  It will return trunk
        # adapters on LPARs, but NOT on VIOS type partitions.  Only CNA has the
        # is_tagged_vlan_supported property; the other types can't be trunk
        # adapters (TODO(IBM) yet?), so always return them.
        return [vif for vif in vif_list if isinstance(vm_wrap, pvm_lpar.LPAR)
                or not getattr(vif, 'is_tagged_vlan_supported', False)]
    except pvm_exc.HttpError as e:
        # If it is a 404 (not found) then just skip.
        if e.response is not None and e.response.status == 404:
            return []
        else:
            raise


def _remove_log_helper(adapter):
    # Remove the log handler from the adapter so we don't log missing VMs
    # Pulling the helpers makes a copy
    helpers = adapter.helpers
    try:
        helpers.remove(pvm_log.log_helper)
    except ValueError:
        # It's not an error if we didn't find it since we don't want it.
        pass
    return helpers


@pvm_retry.retry()
def list_bridges(adapter, host_uuid):
    """
    Queries for the NetworkBridges on the system.  Will return the
    wrapper objects that describe Network Bridges.

    :param adapter: The pypowervm adapter.
    :param host_uuid: The UUID for the host system.
    """
    net_bridges = pvm_net.NetBridge.get(adapter, parent_type=pvm_ms.System,
                                        parent_uuid=host_uuid)

    if len(net_bridges) == 0:
        LOG.warning(_LW('No NetworkBridges detected on the host.'))

    return net_bridges


def device_detail_valid(device_detail, mac, port_id=None):
    """Validate a return from the get_device_details API.

    :param device_detail: The detail dict returned from get_device_details
                          (or one of the list returned from
                          get_devices_details_list).
    :param mac: String MAC address issued to get_device_details, for
                logging.
    :param port_id: UUID of the neutron port.  If None, the port ID is not
                    validated.
    :return: True if the device_detail passes all checks; False otherwise.
    """
    # A device detail will always come back...even if neutron has
    # no idea what the port is.  This WILL happen for PowerVM, maybe
    # an event for the mgmt partition or the secure RMC VIF.  We can
    # detect if Neutron has full device details by simply querying for
    # the mac from the device_detail
    if not device_detail.get('mac_address'):
        LOG.debug("Ignoring VIF with MAC %(mac)s because neutron doesn't know "
                  "about it.", {'mac': mac})
        return False

    if port_id is not None:
        # If the device's id (really the port uuid) doesn't match,
        # ignore it.
        dev_pid = device_detail.get('port_id')
        if dev_pid is None or port_id != dev_pid:
            LOG.warning(_LW(
                "Ignoring port because device_details port_id doesn't match "
                "the port.\nPort: %(port)s\nDevice detail: %(device_detail)s"),
                {'port': port_id, 'device_detail': device_detail})
            return False

    return True



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import mock

from networking_powervm.plugins.ibm.agent.powervm import sea_agent
from networking_powervm.tests.unit.plugins.ibm.powervm import base

from neutron_lib import constants as q_const
from oslo_config import cfg
from pypowervm.wrappers import logical_partition as pvm_lpar
from pypowervm.wrappers import network as pvm_net
from pypowervm.wrappers import virtual_io_server as pvm_vios


def fake_nb(uuid, pvid, tagged_vlans, addl_vlans):
    return mock.MagicMock(
        uuid=uuid,
        load_grps=[mock.MagicMock(pvid=pvid, tagged_vlans=tagged_vlans)],
        list_vlans=mock.Mock(return_value=[pvid] + tagged_vlans + addl_vlans))


class FakeException(Exception):
    """Used to indicate an error in an API the agent calls."""
    pass


class SEAAgentTest(base.BasePVMTestCase):

    def setUp(self):
        super(SEAAgentTest, self).setUp()

        self.agtfx = self.useFixture(base.AgentFx())

        # Mock the mgmt uuid
        self.useFixture(fixtures.MockPatch(
            'pypowervm.tasks.partition.get_mgmt_partition')
        ).mock.return_value = mock.MagicMock(uuid='mgmt_uuid')

        self.mock_parse_sea_mappings = self.useFixture(fixtures.MockPatch(
            'networking_powervm.plugins.ibm.agent.powervm.utils.'
            'parse_sea_mappings')).mock
        self.mock_parse_sea_mappings.return_value = {'default': 'nb_uuid'}

        cfg.CONF.set_override('bridge_mappings', 'the_bridge_maps',
                              group='AGENT')

        self.agent = sea_agent.SharedEthernetNeutronAgent()

    def test_init(self):
        """Verifies the integrity of the agent after being initialized."""
        self.assertEqual('networking-powervm-sharedethernet-agent',
                         self.agent.agent_state.get('binary'))
        self.assertEqual(q_const.L2_AGENT_TOPIC,
                         self.agent.agent_state.get('topic'))
        self.mock_parse_sea_mappings.assert_called_once_with(
            self.agent.adapter, self.agent.host_uuid, 'the_bridge_maps')
        self.assertEqual(
            {'default': 'nb_uuid'},
            self.agent.agent_state['configurations']['bridge_mappings'])
        self.assertEqual('PowerVM Shared Ethernet agent',
                         self.agent.agent_state.get('agent_type'))
        self.assertEqual(True, self.agent.agent_state.get('start_flag'))
        # Other @propertys
        self.assertEqual('sea-agent-%s' % cfg.CONF.host, self.agent.agent_id)
        self.assertEqual(pvm_net.CNA, self.agent.vif_wrapper_class)

    @mock.patch('pypowervm.tasks.network_bridger.ensure_vlans_on_nb')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.provision_devices')
    def test_provision_devices(self, mock_base_prov, mock_ensure):
        """Validates that the provision is invoked with batched VLANs."""
        preq1 = base.mk_preq('plug', 'aa', segment_id=20,
                             phys_network='default', vif_type='pvm_sea')
        preq2 = base.mk_preq('plug', 'bb', segment_id=22,
                             phys_network='default', vif_type='pvm_sea')
        preq3 = base.mk_preq('unplug', 'cc', segment_id=24,
                             phys_network='default', vif_type='pvm_sea')
        # Invoke
        self.agent.provision_devices({preq1, preq2, preq3})

        # Validate that both VLANs are in one call
        mock_ensure.assert_called_once_with(
            self.agent.adapter, self.agent.host_uuid, 'nb_uuid', {20, 22})

        # Validate that the devices were marked up
        mock_base_prov.assert_called_once_with({preq1, preq2})

        # Validate the behavior of a failed VLAN provision.
        mock_ensure.reset_mock()
        mock_base_prov.reset_mock()
        # Have the ensure throw some exception
        mock_ensure.side_effect = FakeException()

        # Invoke
        self.assertRaises(FakeException, self.agent.provision_devices,
                          {preq1, preq2, preq3})

        # Validate that both VLANs are in one call.  Should still occur even
        # though no exception.
        mock_ensure.assert_called_once_with(
            self.agent.adapter, self.agent.host_uuid, 'nb_uuid', {20, 22})

        # However, the port update should not be invoked.
        mock_base_prov.assert_not_called()

    @mock.patch('pypowervm.tasks.network_bridger.remove_vlan_from_nb')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.sea_agent.'
                'SharedEthernetNeutronAgent._get_nb_and_vlan')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.sea_agent.'
                'SharedEthernetNeutronAgent.provision_devices')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'get_vswitch_map')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.list_vifs')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_bridges')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'find_nb_for_cna')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.prov_req.'
                'ProvisionRequest.for_wrappers')
    def test_heal_and_optimize(
            self, mock_preq, mock_find_nb_for_cna, mock_list_bridges,
            mock_list_cnas, mock_vs_map, mock_prov_devs, mock_get_nb_and_vlan,
            mock_nbr_remove):
        """Validates the heal and optimization code.  Limited to 3 deletes."""
        # Fake adapters already on system.
        mgmt_lpar = mock.Mock(spec=pvm_lpar.LPAR, is_mgmt_partition=True)
        reg_lpar = mock.Mock(spec=pvm_lpar.LPAR, is_mgmt_partition=False)
        mgmt_vios = mock.Mock(spec=pvm_vios.VIOS, is_mgmt_partition=True)
        reg_vios = mock.Mock(spec=pvm_vios.VIOS, is_mgmt_partition=False)
        cna1 = mock.MagicMock(mac='00', pvid=30, tagged_vlans=[])
        cna2 = mock.MagicMock(mac='11', pvid=31, tagged_vlans=[32, 33, 34])
        mock_list_cnas.return_value = {
            mgmt_lpar: [cna1], reg_lpar: [cna2], mgmt_vios: [], reg_vios: []}

        # The neutron data.  These will be 'ensured' on the bridge.
        preq1 = base.mk_preq('plug', '00', segment_id=20,
                             phys_network='default')
        preq2 = base.mk_preq('plug', '22', segment_id=22,
                             phys_network='default')
        preq3 = base.mk_preq('unplug', '55', segment_id=55,
                             phys_network='default')
        mock_preq.return_value = [preq1, preq2, preq3]

        # Mock a provision request
        mock_get_nb_and_vlan.return_value = ('nb2_uuid', 23)

        # Mock up network bridges.  VLANs 44, 45, and 46 should be deleted
        # as they are not required by anything.  VLAN 47 should be needed
        # as it is in the pending list.  VLAN 48 should be deleted, but will
        # put over the three delete max count (and therefore would be hit in
        # next pass)
        mock_nb1 = fake_nb('nb_uuid', 20, [], [])
        mock_nb2 = fake_nb('nb2_uuid', 40, [41, 42, 43], [44, 45, 46, 47, 48])
        mock_list_bridges.return_value = [mock_nb1, mock_nb2]
        mock_find_nb_for_cna.return_value = mock_nb2

        # Invoke
        self.agent.heal_and_optimize()

        mock_list_cnas.assert_called_once_with(self.agent.adapter, pvm_net.CNA,
                                               include_vios_and_mgmt=True)
        # Filtered down to the non-mgmt LPAR
        mock_preq.assert_called_once_with(self.agent, {reg_lpar: [cna2]},
                                          'plug')
        mock_list_bridges.assert_called_once_with(self.agent.adapter,
                                                  self.agent.host_uuid)
        mock_prov_devs.assert_called_with([preq1, preq2, preq3])
        mock_get_nb_and_vlan.assert_has_calls(
            [mock.call(req.rpc_device, emit_warnings=False) for req in
             (preq1, preq2, preq3)])
        mock_vs_map.assert_called_once_with(self.agent.adapter,
                                            self.agent.host_uuid)
        mock_find_nb_for_cna.assert_has_calls(
            [mock.call(mock_list_bridges.return_value, cna,
                       mock_vs_map.return_value) for cna in (cna1, cna2)],
            any_order=True)

        # One remove call per net bridge.
        mock_nbr_remove.assert_has_calls(
            [mock.call(
                self.agent.adapter, self.agent.host_uuid, 'nb2_uuid', vlan)
             for vlan in (44, 45, 48)], any_order=True)

        # Validate no remove.
        mock_nbr_remove.reset_mock()
        mock_prov_devs.reset_mock()
        # Set that we can't do the clean up
        cfg.CONF.set_override('automated_powervm_vlan_cleanup', False,
                              group='AGENT')

        # Invoke
        self.agent.heal_and_optimize()

        # Verify.  One ensure call per net bridge.  Zero for the remove as that
        # has been flagged to not clean up.
        mock_nbr_remove.assert_not_called()
        mock_prov_devs.assert_called_with([preq1, preq2, preq3])

    def test_get_nb_and_vlan(self):
        """Be sure nb uuid and vlan parsed from dev properly."""
        self.assertEqual(('nb_uuid', 100), self.agent._get_nb_and_vlan(
            {'physical_network': 'default', 'segmentation_id': 100}))



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron.common import topics
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2.drivers import mech_agent
from neutron.plugins.ml2 import rpc
from neutron_lib.api.definitions import portbindings
from oslo_log import log

LOG = log.getLogger(__name__)


class PvmMechanismDriverBase(mech_agent.SimpleAgentMechanismDriverBase):
    """Base class for PowerVM mechanism drivers."""

    def __init__(self, agent_type, vif_type, **kwargs):
        super(PvmMechanismDriverBase, self).__init__(
            agent_type, vif_type, {portbindings.CAP_PORT_FILTER: False},
            **kwargs)
        self.rpc_publisher = rpc.AgentNotifierApi(topics.AGENT)

    def try_to_bind_segment_for_agent(self, context, segment, agent):
        """Perform binding operation with agent validation.

        :param context: Context of this transaction.
        :param segment: Neutron network object.
        :param agent: Agent configuration.
        :return bindable: Boolean value if port can be bound.
        """
        # This override is a near-duplicate of the superclass impl - but we
        # need to customize the vif_details.
        if self.check_segment_for_agent(segment, agent):
            context.set_binding(
                segment[api.ID], self.vif_type,
                self.customize_vif_details(context, segment, agent))
            return True

        return False

    def customize_vif_details(self, context, segment, agent):
        """Enhance vif details with any driver impl-specific data.

        :param context: Context of this transaction.
        :param segment: Neutron network object.
        :param agent: Agent configuration.
        :return: A dict containing VIF details.  The base impl just invokes
                 _get_vif_details.
        """
        return self._get_vif_details(segment)

    def get_mappings(self, agent):
        """Get bridge mappings from agent configuration.

        Returns the dict of bridge mappings set by the BasePVMNeutronAgent
        subclass corresponding to this mechanism driver.

        :param agent: BasePVMNeutronAgent subclass instance.
        :return mappings: Mappings provided by corresponding agent.  The format
                          will be specific to the agent - see the agent's
                          parse_bridge_mappings method.
        """
        return agent['configurations'].get('bridge_mappings', {})

    def _get_vif_details(self, segment):
        """Gets customized vif details for this mechanism driver.

        :param segment: Neutron network object.
        :return: A dict containing VIF details, including:
                 - vlan: (Integer) VLAN ID for the vif.  None for flat VLANs.
        """
        vif_details = copy.copy(self.vif_details)
        vlan_id = segment.get(api.SEGMENTATION_ID)
        vif_details[portbindings.VIF_DETAILS_VLAN] = (
            str(vlan_id) if vlan_id is not None else None)
        return vif_details



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from pypowervm.wrappers import iocard as pvm_card

import mock

from networking_powervm.plugins.ibm.agent.powervm import sriov_agent
from networking_powervm.tests.unit.plugins.ibm.powervm import base


class SRIOVAgentTest(base.BasePVMTestCase):
    def setUp(self):
        super(SRIOVAgentTest, self).setUp()
        self.agtfx = self.useFixture(base.AgentFx())
        sriov_adaps = [
            mock.Mock(phys_ports=[
                mock.Mock(loc_code='loc1', label='foo'),
                mock.Mock(loc_code='loc2', label='')]),
            mock.Mock(phys_ports=[
                mock.Mock(loc_code='loc3', label='bar'),
                mock.Mock(loc_code='loc4', label='foo')])]
        self.agtfx.sys = mock.Mock(
            asio_config=mock.Mock(sriov_adapters=sriov_adaps))
        self.agtfx.sysget.return_value = [self.agtfx.sys]
        self.agtfx.sys.refresh.return_value = self.agtfx.sys
        self.agt = sriov_agent.SRIOVNeutronAgent()

    def test_init(self):
        """Test __init__, customize_agent_state, and parse_bridge_mappings."""
        # SR-IOV-specific agent state attrs were set
        self.assertEqual(
            'networking-powervm-sriov-agent', self.agt.agent_state['binary'])
        self.assertEqual('PowerVM SR-IOV Ethernet agent',
                         self.agt.agent_state['agent_type'])
        # Validate customize_agent_state
        self.assertEqual(
            2, self.agt.agent_state['configurations']['default_redundancy'])
        self.assertIsNone(
            self.agt.agent_state['configurations']['default_capacity'])
        # Validate parse_bridge_mappings
        self.agtfx.sys.refresh.assert_called_once_with()
        br_map = self.agt.br_map
        self.assertEqual({'default', 'foo', 'bar'}, set(br_map.keys()))
        self.assertEqual(['loc2'], br_map['default'])
        self.assertEqual(['loc3'], br_map['bar'])
        self.assertEqual({'loc1', 'loc4'}, set(br_map['foo']))

        # Ensure conf.vnic_required_vfs affects default_redundancy & _capacity
        cfg.CONF.set_override('vnic_required_vfs', 3, group='AGENT')
        cfg.CONF.set_override('vnic_vf_capacity', 0.06, group='AGENT')
        agt = sriov_agent.SRIOVNeutronAgent()
        self.assertEqual(
            3, agt.agent_state['configurations']['default_redundancy'])
        self.assertEqual(
            0.06, agt.agent_state['configurations']['default_capacity'])

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.sriov_agent.'
                'SRIOVNeutronAgent._refresh_bridge_mappings_to_neutron')
    def test_port_update(self, mock_refresh):
        """Test the port_update override."""
        self.agt.port_update('context')
        mock_refresh.assert_called_once_with()

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.sriov_agent.'
                'SRIOVNeutronAgent.parse_bridge_mappings')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent._report_state')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.list_vifs')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.prov_req.'
                'ProvisionRequest.for_wrappers')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.provision_devices')
    def test_heal_and_optimize(self, mock_prv, mock_preq, mock_vifs, mock_rpt,
                               mock_pbm):
        """Test heal_and_optimize and _refresh_bridge_mappings_to_neutron."""
        self.agt.heal_and_optimize()
        mock_pbm.assert_called_once_with()
        mock_rpt.assert_called_once_with()
        self.assertEqual(
            mock_pbm.return_value,
            self.agt.agent_state['configurations']['bridge_mappings'])
        mock_vifs.assert_called_once_with(self.agt.adapter, pvm_card.VNIC)
        mock_preq.assert_called_once_with(self.agt, mock_vifs.return_value,
                                          'plug')
        mock_prv.assert_called_once_with(mock_preq.return_value)

    def test_is_hao_event(self):
        self.assertFalse(self.agt.is_hao_event(mock.Mock(detail=None)))
        self.assertFalse(self.agt.is_hao_event(mock.Mock(detail='')))
        self.assertFalse(self.agt.is_hao_event(mock.Mock(detail='Foo')))
        self.assertFalse(self.agt.is_hao_event(mock.Mock(detail='Foo,Bar')))
        self.assertTrue(self.agt.is_hao_event(mock.Mock(
            detail='SRIOVPhysicalPort.ConfigChange')))
        self.assertTrue(self.agt.is_hao_event(mock.Mock(
            detail='Foo,SRIOVPhysicalPort.ConfigChange')))
        self.assertTrue(self.agt.is_hao_event(mock.Mock(
            detail='SRIOVPhysicalPort.ConfigChange,Bar')))
        self.assertTrue(self.agt.is_hao_event(mock.Mock(
            detail='Foo,SRIOVPhysicalPort.ConfigChange,Bar')))



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.hacking import checks


def no_log_warn(logical_line):
    """Disallow 'LOG.warn('

    """
    if logical_line.startswith('LOG.warn('):
        yield(0, 'P301 Use LOG.warning() rather than LOG.warn()')


def no_os_popen(logical_line):
    """Disallow 'os.popen('

    Deprecated library function os.popen() Replace it using subprocess
    https://bugs.launchpad.net/tempest/+bug/1529836

    P302
    """

    if 'os.popen(' in logical_line:
        yield(0, 'P302 Deprecated library function os.popen(). '
                 'Replace it using subprocess module. ')


def factory(register):
    register(no_log_warn)
    register(no_os_popen)
    checks.factory(register)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_config import cfg
from pypowervm.helpers import log_helper as log_hlp
from pypowervm.helpers import vios_busy as vio_hlp

from networking_powervm.plugins.ibm.agent.powervm import agent_base
from networking_powervm.tests.unit.plugins.ibm.powervm import base


def FakeNPort(mac, segment_id, phys_network):
    return {'mac_address': mac, 'segmentation_id': segment_id,
            'physical_network': phys_network}


class FakeAgent(agent_base.BasePVMNeutronAgent):
    def __init__(self):
        self.mock_vif_wrapper_class = mock.Mock()
        self.parse_bridge_mappings = mock.Mock()
        self.heal_and_optimize = mock.Mock()
        self.customize_agent_state = mock.Mock()
        self.setup_adapter_called = mock.Mock()
        super(FakeAgent, self).__init__()

    @property
    def agent_id(self):
        return 'agent_id'

    @property
    def agent_binary_name(self):
        return 'agent_binary_name'

    @property
    def agent_type(self):
        return 'agent_type'

    @property
    def vif_wrapper_class(self):
        return self.mock_vif_wrapper_class

    def setup_adapter(self):
        super(FakeAgent, self).setup_adapter()
        self.setup_adapter_called()


class TestAgentBaseInit(base.BasePVMTestCase):
    """A test class to validate the set up of the agent with the API.

    This is typically mocked out in fixtures otherwise.
    """
    def setUp(self):
        super(TestAgentBaseInit, self).setUp()
        self.agtfx = self.useFixture(base.AgentFx())
        # For init
        self.adpt = self.agtfx.adpt
        self.sess = self.agtfx.sess
        self.adpt.return_value.session = self.sess.return_value
        self.sysget = self.agtfx.sysget
        self.sys = self.agtfx.sys
        # For RPC
        self.plg_rpt_st_api = self.agtfx.plg_rpt_st_api
        self.gacwos = self.agtfx.gacwos
        self.crt_cons = self.agtfx.crt_cons
        self.filc = self.agtfx.filc
        # PluginAPI
        self.plug_api = self.agtfx.plug_api
        # VIFEventHandler
        self.veh = self.agtfx.veh
        # An agent to use
        self.agt = FakeAgent()

    def test_init(self):
        """Test BasePVMNeutronAgent __init__, setup_rpc, PVMPluginAPI."""
        # __init__ part 1
        self.sess.assert_called_once_with(conn_tries=300)
        self.adpt.assert_called_once_with(
            self.sess.return_value, helpers=[log_hlp.log_helper,
                                             vio_hlp.vios_busy_retry_helper])
        self.assertEqual(self.adpt.return_value, self.agt.adapter)
        # setup_adapter override was invoked
        self.agt.setup_adapter_called.assert_called_once_with()
        self.sysget.assert_called_once_with(self.adpt.return_value)
        self.assertEqual(self.sys, self.agt.msys)
        self.assertEqual(self.sys.uuid, self.agt.host_uuid)
        self.agt.parse_bridge_mappings.assert_called_once_with()
        self.assertEqual(self.agt.parse_bridge_mappings.return_value,
                         self.agt.br_map)
        self.assertEqual({
            'binary': 'agent_binary_name', 'host': cfg.CONF.host,
            'topic': 'N/A',
            'configurations': {'bridge_mappings': self.agt .br_map},
            'agent_type': 'agent_type', 'start_flag': True},
            self.agt.agent_state)
        self.agt.customize_agent_state.assert_called_once_with()
        # _setup_rpc
        self.assertEqual('q-agent-notifier', self.agt.topic)
        self.assertIsInstance(self.agt.plugin_rpc, agent_base.PVMPluginApi)
        self.plug_api.assert_called_once_with('q-plugin')
        self.plg_rpt_st_api.assert_called_once_with('q-plugin')
        self.assertEqual(self.plg_rpt_st_api.return_value, self.agt.state_rpc)
        self.gacwos.assert_called_once_with()
        self.assertEqual(self.gacwos.return_value, self.agt.context)
        self.crt_cons.assert_called_once_with(
            [self.agt], 'q-agent-notifier', [['port', 'update']])
        self.assertEqual(self.crt_cons.return_value, self.agt.connection)
        self.filc.assert_called_once_with(self.agt._report_state)
        self.filc.return_value.start.assert_called_once_with(interval=30.0)
        # __init__ part 2
        self.sess.return_value.get_event_listener.assert_called_once_with()
        self.veh.assert_called_once_with(self.agt)
        self.assertEqual(self.agt._vif_event_handler, self.veh.return_value)
        mock_evt_list = self.sess.return_value.get_event_listener.return_value
        mock_evt_list.subscribe.assert_called_once_with(
            self.agt._vif_event_handler)

        # No report interval => No looping call
        self.filc.reset_mock()
        self.filc.return_value.start.reset_mock()
        cfg.CONF.set_override('report_interval', 0, group='AGENT')
        FakeAgent()
        self.filc.assert_not_called()
        self.filc.return_value.start.assert_not_called()

    def test_report_state(self):
        self.agt._report_state()
        self.assertEqual(0, self.agt.agent_state['configurations']['devices'])
        self.plg_rpt_st_api.return_value.report_state.assert_called_once_with(
            self.agt.context, self.agt.agent_state)
        self.assertNotIn('start_flag', self.agt.agent_state)

    def test_update_device_up(self):
        self.agt.plugin_rpc.update_device_up = mock.Mock()
        self.agt.update_device_up({'device': 'the_device'})
        self.agt.plugin_rpc.update_device_up.assert_called_once_with(
            self.agt.context, 'the_device', self.agt.agent_id, cfg.CONF.host)

    def test_update_device_down(self):
        self.agt.plugin_rpc.update_device_down = mock.Mock()
        self.agt.update_device_down({'device': 'the_device'})
        self.agt.plugin_rpc.update_device_down.assert_called_once_with(
            self.agt.context, 'the_device', self.agt.agent_id, cfg.CONF.host)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.norm_mac')
    def test_get_device_details(self, mock_mac):
        """Test get_device_details and get_devices_details_list."""
        # get_device_details
        self.agt.plugin_rpc.get_device_details = mock.Mock()
        self.agt.get_device_details('mac')
        mock_mac.assert_called_once_with('mac')
        self.agt.plugin_rpc.get_device_details.assert_called_once_with(
            self.agt.context, mock_mac.return_value, self.agt.agent_id)
        # get_devices_details_list
        mock_mac.reset_mock()
        mock_mac.side_effect = ['mac1', 'mac2', 'mac3']
        self.agt.plugin_rpc.get_devices_details_list = mock.Mock()
        self.agt.get_devices_details_list([1, 2, 3])
        mock_mac.assert_has_calls([mock.call(val) for val in [1, 2, 3]])
        self.agt.plugin_rpc.get_devices_details_list.assert_called_once_with(
            self.agt.context, ['mac1', 'mac2', 'mac3'], self.agt.agent_id)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.update_device_up')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'BasePVMNeutronAgent.update_device_down')
    def test_provision_devices(self, mock_down, mock_up):
        req1 = base.mk_preq('plug', 'mac1')
        req2 = base.mk_preq('unplug', 'mac2')
        req3 = base.mk_preq('plug', 'mac3')
        req4 = base.mk_preq('frobnicate', 'mac4')
        self.agt.provision_devices((req1, req2, req3, req4))
        mock_up.assert_has_calls([mock.call(req.rpc_device) for req in
                                  (req1, req3)])
        mock_down.assert_called_once_with(req2.rpc_device)

    @mock.patch('time.sleep')
    def test_rpc_loop(self, mock_sleep):
        # This is going to be a little weird.  To break out of the while True
        # loop, we'll have to have the exception path raise a new exception.
        # Set this up as follows:
        # 1) Run heal_and_optimize successfully
        # 2) Run heal_and_optimize successfully
        # 3) heal_and_optimize raises to prove the exception path
        # 4) Run heal_and_optimize successfully
        # 5) heal_and_optimize raises to allow us to bail
        self.agt.heal_and_optimize.side_effect = [
            None, None, Exception, None, Exception]
        # Hence time.sleep should raise on the fifth invocation
        mock_sleep.side_effect = [None, None, None, None, KeyboardInterrupt]
        # Run it
        self.assertRaises(KeyboardInterrupt, self.agt.rpc_loop)
        self.agt.heal_and_optimize.assert_has_calls([mock.call()] * 5)
        mock_sleep.assert_has_calls([mock.call(val) for val in
                                     (1800, 1800, 5, 1800, 5)])


class TestVIFEventHandler(base.BasePVMTestCase):
    """Validates that the VIFEventHandler can be invoked properly."""

    def setUp(self):
        super(TestVIFEventHandler, self).setUp()

        self.mock_agent = mock.MagicMock()
        self.handler = agent_base.VIFEventHandler(self.mock_agent)

    def test_init(self):
        """Proper initialization of instance attributes."""
        self.assertEqual(self.mock_agent, self.handler.agent)
        self.assertEqual(self.mock_agent.adapter, self.handler.adapter)
        self.assertTrue(self.handler.just_started)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.list_vifs')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.prov_req.'
                'ProvisionRequest.for_wrappers')
    def test_refetch_all(self, mock_preqs, mock_vifs):
        prov_req_set = set()
        req1 = base.mk_preq('plug', 'mac1', lpar_uuid='lpar1')
        req2 = base.mk_preq('unplug', 'mac2', lpar_uuid='lpar2')
        req3 = base.mk_preq('some_other_action', 'mac3', lpar_uuid='lpar3')
        req4 = base.mk_preq('plug', 'mac4', lpar_uuid='lpar4')

        # 1) No-op
        mock_preqs.return_value = []
        self.handler._refetch_all(prov_req_set)
        mock_vifs.assert_called_once_with(
            self.handler.adapter, self.mock_agent.vif_wrapper_class)
        mock_preqs.assert_called_once_with(
            self.mock_agent, mock_vifs.return_value, 'plug')
        self.assertEqual(set(), prov_req_set)

        # 2) Add some reqs. (IRL, no 'unplug' reqs would come in here, but this
        #    sets up for the next test.)
        mock_preqs.return_value = [req1, req2, req3]
        self.handler._refetch_all(prov_req_set)
        # All reqs were added.
        self.assertEqual(set(mock_preqs.return_value), prov_req_set)

        # 3) a) Trim plug reqs (req1 goes away; req3 stays);
        #    b) add new reqs (req4);
        #    c) consolidate (req2 appears once).
        mock_preqs.return_value = [req2, req4]
        self.handler._refetch_all(prov_req_set)
        self.assertEqual({req2, req3, req4}, prov_req_set)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.prov_req.'
                'ProvisionRequest.for_event')
    def test_process_event(self, mock_preq):
        req1 = base.mk_preq('plug', 'mac1', lpar_uuid='lpar1')
        req2 = base.mk_preq('unplug', 'mac2', lpar_uuid='lpar2')
        req3 = base.mk_preq('some_other_action', 'mac3', lpar_uuid='lpar3')
        # "matches" req2
        req4 = base.mk_preq('plug', 'mac2', lpar_uuid='lpar2')
        # Seed the req set
        prov_req_set = {req1, req2, req3}

        # 1) No req returned => no-op
        mock_preq.return_value = None
        self.handler._process_event('event', prov_req_set)
        mock_preq.assert_called_once_with(self.mock_agent, 'event')
        self.assertEqual({req1, req2, req3}, prov_req_set)

        # 2) Req req4 returned:
        #    a) req4 added to the set;
        #    b) "matching" reqs (req2) removed.
        mock_preq.return_value = req4
        self.handler._process_event('event', prov_req_set)
        self.assertEqual({req1, req3, req4}, prov_req_set)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'VIFEventHandler._refetch_all')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.agent_base.'
                'VIFEventHandler._process_event')
    def test_process(self, mock_proc_evt, mock_ref_all):
        def _process_event(event, prs):
            prs.add('event %s' % event.etype)
        mock_proc_evt.side_effect = _process_event

        def _refetch_all(prs):
            prs.add('refetch')
        mock_ref_all.side_effect = _refetch_all

        # Refetch events
        r_evts = [mock.Mock(etype=val, detail='One,Two') for val in
                  agent_base.FULL_REFETCH_EVENTS]
        # Single object events
        s_evts = [mock.Mock(etype=val) for val in agent_base.SINGLE_OBJ_EVENTS]
        # Ignored events
        i_evts = [mock.Mock(etype=val, detail='Three') for val in
                  ('foo', 'bar')]

        # 1) No events => no action, just_started stays True
        self.assertTrue(self.handler.just_started)
        self.handler.process([])
        mock_ref_all.assert_not_called()
        mock_proc_evt.assert_not_called()
        self.mock_agent.is_hao_event.assert_not_called()
        self.mock_agent.provision_devices.assert_called_once_with(set())
        self.assertTrue(self.handler.just_started)

        self.mock_agent.provision_devices.reset_mock()

        # 2) Ignorable followed by non-ignorable:
        #    a) Processors are called;
        #    b) Non-ignorables are provisioned.
        #    c) No heal-and-optimize events.
        self.mock_agent.is_hao_event.return_value = False
        evts = [i_evts[0], r_evts[0], s_evts[0]]
        self.handler.process(evts)
        mock_ref_all.assert_called_once_with(mock.ANY)
        mock_proc_evt.assert_called_once_with(s_evts[0], mock.ANY)
        self.mock_agent.is_hao_event.assert_has_calls(
            [mock.call(evt) for evt in evts])
        self.mock_agent.heal_and_optimize.assert_not_called()
        self.mock_agent.provision_devices.assert_called_once_with(
            {'refetch', 'event INVALID_URI'})
        self.assertFalse(self.handler.just_started)

        mock_ref_all.reset_mock()
        mock_proc_evt.reset_mock()
        self.mock_agent.provision_devices.reset_mock()

        # 3) a) Cover all events;
        #    b) Non-ignorables are provisioned.
        #    c) Some heal-and-optimize events.
        evts = s_evts + i_evts + r_evts
        # Make is_hao_event return True periodically
        self.mock_agent.is_hao_event.side_effect = map(
            lambda x: False if x % 3 else True, range(len(evts)))
        self.handler.process(evts)
        mock_ref_all.assert_has_calls([mock.call(mock.ANY)] * len(r_evts))
        mock_proc_evt.assert_has_calls([mock.call(evt, mock.ANY)
                                        for evt in s_evts])
        # is_hao_event was called for every event
        self.mock_agent.is_hao_event.assert_has_calls(
            [mock.call(evt) for evt in evts])
        # heal_and_optimize was only called once - the loop accumulates
        self.mock_agent.heal_and_optimize.assert_called_once_with()
        self.mock_agent.provision_devices.assert_called_once_with(
            {'refetch'} | {'event %s' % val for val in
                           agent_base.SINGLE_OBJ_EVENTS})



# Copyright 2014, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.plugins.common import constants as p_constants
from neutron_lib.api.definitions import portbindings
from oslo_log import log

from networking_powervm.plugins.ibm.agent.powervm import constants as pconst
from networking_powervm.plugins.ml2.drivers import mech_pvm_base

LOG = log.getLogger(__name__)


class PvmSRIOVMechanismDriver(mech_pvm_base.PvmMechanismDriverBase):
    """Attach to networks using PowerVM SR-IOV agent.

    The PvmSRIOVMechanismDriver integrates the ml2 plugin with the
    PowerVM SRIOV Agent.
    """

    def __init__(self):
        super(PvmSRIOVMechanismDriver, self).__init__(
            pconst.AGENT_TYPE_PVM_SRIOV, pconst.VIF_TYPE_PVM_SRIOV,
            supported_vnic_types=[portbindings.VNIC_DIRECT])

    def get_allowed_network_types(self, agent=None):
        """Get a list of allowed network types.

        This mechanism driver supports only VLAN and FLAT network types.

        :param agent: Not used
        :return: List of neutron.plugins.common.constants.TYPE_* strings.
        """
        return [p_constants.TYPE_FLAT, p_constants.TYPE_VLAN]

    def customize_vif_details(self, context, segment, agent):
        """Gets customized vif details for this mechanism driver.

        :param context: Context of this transaction.
        :param segment: Neutron network object.
        :param agent: SR-IOV vNIC agent
        :return: A dict containing VIF details, including the following keys:
                 - vlan: (Integer) VLAN ID for the vif.
                 - physical_ports: (List(str)) List of physical location codes
                   of SR-IOV physical ports cabled to the neutron network for
                   this vif (based on matching the port labels to the network
                   name).
                 - default_redundancy: (Integer) Default number of VFs with
                   which to back new vNICs.  May be overridden by
                   binding:profile['vnic_required_vfs'] at plug time.
        """
        vif_details = (
            super(PvmSRIOVMechanismDriver, self).customize_vif_details(
                context, segment, agent))
        vif_details['physical_ports'] = self.get_mappings(agent).get(
            segment['physical_network'], [])
        vif_details['physical_network'] = segment['physical_network']
        profile = context.current.get(portbindings.PROFILE, {})
        # TODO(efried): binding:profile info is not in the 'profile' var!
        # Redundancy: from binding:profile or the ml2 conf.
        vif_details['redundancy'] = int(profile.get(
            'vnic_required_vfs',
            agent['configurations']['default_redundancy']))
        # Capacity: from binding:profile or the ml2 conf.  If unspecified in
        # either, let the platform default.
        cap = profile.get(
            'capacity', agent['configurations']['default_capacity'])
        try:
            vif_details['capacity'] = float(cap)
        except (TypeError, ValueError):
            # cap may be None or 'None' at this point, depending on the source.
            vif_details['capacity'] = None

        return vif_details



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_powervm.plugins.ibm.agent.powervm import exceptions as np_exc
from networking_powervm.plugins.ibm.agent.powervm import utils
from networking_powervm.tests.unit.plugins.ibm.powervm import base

from pypowervm.wrappers import logical_partition as pvm_lpar
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import virtual_io_server as pvm_vios


class UtilsTest(base.BasePVMTestCase):
    """Tests the utility functions for the Shared Ethernet Adapter Logic."""

    def __cna(self, mac):
        """Create a Client Network Adapter mock."""

        class FakeCNA(object):

            @property
            def slot(self):
                return 1

            @property
            def mac(self):
                return mac

            @property
            def pvid(self):
                return 1

        return FakeCNA()

    def test_norm_mac(self):
        EXPECTED = "12:34:56:78:90:ab"
        self.assertEqual(EXPECTED, utils.norm_mac("12:34:56:78:90:ab"))
        self.assertEqual(EXPECTED, utils.norm_mac("1234567890ab"))
        self.assertEqual(EXPECTED, utils.norm_mac("12:34:56:78:90:AB"))
        self.assertEqual(EXPECTED, utils.norm_mac("1234567890AB"))

    @mock.patch('pypowervm.wrappers.network.NetBridge.get')
    def test_list_bridges(self, mock_nbrget):
        """Test that we can load the bridges in properly."""
        mock_nbrget.return_value = ['br1', 'br2']

        # Assert that two are read in
        bridges = utils.list_bridges('adpt', 'host_uuid')
        self.assertEqual(2, len(bridges))
        mock_nbrget.assert_called_once_with(
            'adpt', parent_type=pvm_ms.System, parent_uuid='host_uuid')

    @mock.patch('pypowervm.wrappers.network.VSwitch.get')
    def test_get_vswitch_map(self, mock_get):
        # Create mocks
        mock_get.return_value = [
            mock.Mock(related_href='http://9.1.2.3/test', switch_id=0),
            mock.Mock(related_href='http://9.4.5.6/test', switch_id=5)]

        # Run test and verify results
        self.assertEqual({0: 'http://9.1.2.3/test', 5: 'http://9.4.5.6/test'},
                         utils.get_vswitch_map('adpt', 'host_uuid'))
        mock_get.assert_called_once_with(
            'adpt', parent_type=pvm_ms.System, parent_uuid='host_uuid')

    def test_find_nb_for_cna(self):
        # Create mocks
        nb_wrap = mock.Mock()
        nb_wrap.vswitch_id = '0'
        nb_wrap.supports_vlan = mock.Mock(return_value=True)
        nb_wraps = [nb_wrap]

        vswitch_map = {'0': 'http://9.1.2.3/vs1',
                       '1': 'http://9.1.2.3/vs2'}

        mock_client_adpt = mock.MagicMock(vswitch_uri='http://9.1.2.3/vs1')

        # Should have a proper URI, so it should match
        resp = utils.find_nb_for_cna(nb_wraps, mock_client_adpt, vswitch_map)
        self.assertIsNotNone(resp)

        # Should not match if we change the vswitch URI
        mock_client_adpt.vswitch_uri = "Fake"
        resp = utils.find_nb_for_cna(nb_wraps, mock_client_adpt, vswitch_map)
        self.assertIsNone(resp)

    @mock.patch('pypowervm.tasks.partition.get_partitions')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                '_find_vifs')
    def test_list_vifs(self, mock_find_vifs, mock_get_partitions):
        pars = ['par1', 'par2', 'par3']
        mock_get_partitions.return_value = pars
        mock_find_vifs.side_effect = ['vif1', 'vif2', 'vif3']

        # Default (no VIOS/mgmt)
        self.assertEqual({'par1': 'vif1', 'par2': 'vif2', 'par3': 'vif3'},
                         utils.list_vifs('adap', 'vif_class'))
        mock_get_partitions.assert_called_once_with(
            'adap', lpars=True, vioses=False, mgmt=False)
        mock_find_vifs.assert_has_calls(
            [mock.call('adap', 'vif_class', vm_wrap) for vm_wrap in pars])

        mock_get_partitions.reset_mock()
        mock_find_vifs.reset_mock()
        mock_find_vifs.side_effect = ['vif1', 'vif2', 'vif3']

        # With VIOS/mgmt
        self.assertEqual({'par1': 'vif1', 'par2': 'vif2', 'par3': 'vif3'},
                         utils.list_vifs('adap', 'vif_class',
                                         include_vios_and_mgmt=True))
        mock_get_partitions.assert_called_once_with(
            'adap', lpars=True, vioses=True, mgmt=True)
        mock_find_vifs.assert_has_calls(
            [mock.call('adap', 'vif_class', vm_wrap) for vm_wrap in pars])

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_bridges')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.get')
    def test_parse_sea_mappings(self, mock_vioget, mock_list_br):
        # Create mocks
        mock_vios = mock.Mock()
        mock_vios.configure_mock(name='21-25D0A',
                                 uuid="4E0B057C-F052-4609-8EDE-071C7FC485BD")

        mock_sea = mock.Mock(
            dev_name='ent8',
            vio_uri='http://localhost:12080/rest/api/uom/ManagedSystem/'
                    'b7f6d2f3-c4f3-33e4-91ea-1aaeada65d7b/VirtualIOServer/'
                    '4E0B057C-F052-4609-8EDE-071C7FC485BD')
        nb_wrap = mock.Mock()
        nb_wrap.seas = [mock_sea]
        nb_wrap.uuid = '764f3423-04c5-3b96-95a3-4764065400bd'
        nb_wraps = [nb_wrap]
        mock_list_br.return_value = nb_wraps
        mock_vioget.return_value = [mock_vios]

        # Run actual test
        resp = utils.parse_sea_mappings('adap', 'host_uuid',
                                        'default:ent8:21-25D0A')

        # Verify results
        self.assertEqual(1, len(resp.keys()))
        self.assertIn('default', resp)
        self.assertEqual('764f3423-04c5-3b96-95a3-4764065400bd',
                         resp['default'])

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_bridges')
    def test_parse_sea_mappings_no_bridges(self, mock_list_br):
        mock_list_br.return_value = []
        self.assertRaises(np_exc.NoNetworkBridges, utils.parse_sea_mappings,
                          None, 'host_uuid', '1:2:3')

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                '_parse_empty_bridge_mapping')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'list_bridges')
    def test_parse_call_to_empty_bridge(self, mock_list_br, mock_empty):
        mock_list_br.return_value = ['br1']

        utils.parse_sea_mappings(None, 'host_uuid', '')

        # Make sure the _parse_empty_bridge_mapping method was called
        self.assertEqual(1, mock_empty.call_count)

    def test_parse_empty_bridge_mappings(self):
        proper_wrap = mock.MagicMock()
        proper_wrap.uuid = '5'
        resp = utils._parse_empty_bridge_mapping([proper_wrap])

        self.assertEqual({'default': '5'}, resp)

        # Try the failure path
        self.assertRaises(np_exc.MultiBridgeNoMapping,
                          utils._parse_empty_bridge_mapping,
                          [proper_wrap, mock.Mock()])

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                '_remove_log_helper')
    def test_find_cnas(self, mock_rmlog):
        mock_vif_class = mock.Mock()
        vea1 = mock.Mock(uuid='1', is_tagged_vlan_supported=True)
        vea2 = mock.Mock(uuid='2', is_tagged_vlan_supported=False)
        vea3 = mock.Mock(uuid='3', is_tagged_vlan_supported=True)
        mock_vif_class.get.return_value = [vea1, vea2, vea3]

        lpar = mock.Mock(spec=pvm_lpar.LPAR)
        vios = mock.Mock(spec=pvm_vios.VIOS)

        # The LPAR type should include the trunk adapters
        self.assertEqual([vea1, vea2, vea3],
                         utils._find_vifs('adap', mock_vif_class, lpar))
        mock_vif_class.get.assert_called_once_with(
            'adap', parent=lpar, helpers=mock_rmlog.return_value)
        mock_rmlog.assert_called_once_with('adap')

        mock_vif_class.get.reset_mock()
        mock_rmlog.reset_mock()

        # The vios type should ignore the trunk adapters
        self.assertEqual([vea2],
                         utils._find_vifs('adap', mock_vif_class, vios))
        mock_vif_class.get.assert_called_once_with(
            'adap', parent=vios, helpers=mock_rmlog.return_value)
        mock_rmlog.assert_called_once_with('adap')



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
import sys

from neutron.common import config as n_config
from neutron.conf.agent import common as a_config
from oslo_config import cfg
from oslo_log import log as logging

from networking_powervm._i18n import _
from networking_powervm._i18n import _LI
from networking_powervm.plugins.ibm.agent.powervm import agent_base
from networking_powervm.plugins.ibm.agent.powervm import constants as p_const
from networking_powervm.plugins.ibm.agent.powervm import prov_req as preq
from networking_powervm.plugins.ibm.agent.powervm import utils
from pypowervm.wrappers import iocard as pvm_card

eventlet.monkey_patch()


LOG = logging.getLogger(__name__)

agent_opts = [
    cfg.IntOpt('vnic_required_vfs', default=2, min=1,
               help=_('Redundancy level for SR-IOV backed vNIC attachments. '
                      'Minimum value is 1.')),
    cfg.FloatOpt('vnic_vf_capacity',
                 help=_("Float up to 4dp between 0.0000 and 1.0000 indicating "
                        "the minimum guaranteed capacity of the VFs backing "
                        "an SR-IOV vNIC.  Must be a multiple of each physical "
                        "port's minimum capacity granularity.  If omitted, "
                        "defaults to the minimum capacity granularity for "
                        "each port."))
]

cfg.CONF.register_opts(agent_opts, "AGENT")
a_config.register_agent_state_opts_helper(cfg.CONF)
a_config.register_root_helper(cfg.CONF)

ACONF = cfg.CONF.AGENT

VIF_TYPE_PVM_SRIOV = 'pvm_sriov'


class SRIOVNeutronAgent(agent_base.BasePVMNeutronAgent):
    """
    Provides VLAN networks for the PowerVM platform that run through
    shared-mode SR-IOV adapters within the Virtual I/O Servers in the form of
    vNIC.  Designed to be compatible with the ML2 Neutron Plugin.
    """
    @property
    def agent_id(self):
        return 'sriov-agent-%s' % cfg.CONF.host

    @property
    def agent_binary_name(self):
        """Name of the executable under which the SR-IOV agent runs."""
        return p_const.AGENT_BIN_SRIOV

    @property
    def agent_type(self):
        return p_const.AGENT_TYPE_PVM_SRIOV

    @property
    def vif_wrapper_class(self):
        return pvm_card.VNIC

    @property
    def vif_type(self):
        return VIF_TYPE_PVM_SRIOV

    def customize_agent_state(self):
        """Set SR-IOV-specific configurations in the agent_state."""
        self.agent_state['configurations']['default_redundancy'] = (
            ACONF.vnic_required_vfs)
        self.agent_state['configurations']['default_capacity'] = (
            ACONF.vnic_vf_capacity)

    def parse_bridge_mappings(self):
        """Dict of {physnet: [physloc, ...]} for SR-IOV physical ports.

        The physical network name is retrieved from the SR-IOV physical port's
        label.  The user is responsible for setting the label prior to agent
        activation.  Unlabeled ports will be assumed to belong to the 'default'
        network.

        :return mapping: Return a mapping of physical network names to lists of
                         SR-IOV physical port location codes.  Example:
                         {'default': ['U78C9.001.WZS094N-P1-C7-T2',
                                      'U78C9.001.WZS094N-P2-C1-T3'],
                          'prod': ['U78C9.001.WZS094N-P2-C7-T2',
                                   'U78C9.001.WZS094N-P1-C1-T3',
                                   'U78C9.001.WZS094N-P1-C3-T1']}
        """
        self.msys = self.msys.refresh()
        mapping = {}
        for sriov in self.msys.asio_config.sriov_adapters:
            for pport_w in sriov.phys_ports:
                label = pport_w.label or 'default'
                if label not in mapping:
                    mapping[label] = []
                mapping[label].append(pport_w.loc_code)
        return mapping

    def _refresh_bridge_mappings_to_neutron(self):
        """Refresh the label:physloc mappings and report to neutron."""
        LOG.debug("Refreshing bridge mappings to neutron.  Before: %s",
                  self.agent_state['configurations']['bridge_mappings'])
        nbm = self.parse_bridge_mappings()
        self.agent_state['configurations']['bridge_mappings'] = nbm
        LOG.debug("After: %s", nbm)
        self._report_state()

    def port_update(self, context, **kwargs):
        """Refresh neutron bridge mappings on port update."""
        self._refresh_bridge_mappings_to_neutron()

    def heal_and_optimize(self):
        """Ensure bridge mappings are current, and all VNICs are marked up. """
        self._refresh_bridge_mappings_to_neutron()

        # Create ProvisionRequests for all VNICs on all (non-management client)
        # partitions...
        lpar_vnic_map = utils.list_vifs(self.adapter, self.vif_wrapper_class)
        prov_reqs = preq.ProvisionRequest.for_wrappers(self, lpar_vnic_map,
                                                       preq.PLUG)
        # ...and mark them 'up' in neutron.
        self.provision_devices(prov_reqs)

    def is_hao_event(self, evt):
        """Determines if an Event warrants a heal_and_optimize.

        :param evt: A pypowervm.wrappers.event.Event wrapper to inspect.
        :return: True if heal_and_optimize should be invoked as a result of
                 this event; False otherwise.
        """
        return evt.detail and 'SRIOVPhysicalPort.ConfigChange' in evt.detail


def main():
    # Read in the command line args
    n_config.init(sys.argv[1:])
    n_config.setup_logging()

    # Build then run the agent
    agent = SRIOVNeutronAgent()
    LOG.info(_LI("PowerVM SR-IOV Agent initialized and running."))
    agent.rpc_loop()


if __name__ == "__main__":
    main()



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import mock
import testtools

from networking_powervm.plugins.ibm.agent.powervm import prov_req


class TestProvisionRequest(testtools.TestCase):
    def setUp(self):
        super(TestProvisionRequest, self).setUp()
        self.mock_time = self.useFixture(fixtures.MockPatch('time.time')).mock

    def test_init(self):
        # This actually works with nothing in the dev (it should probably fail)
        preq = prov_req.ProvisionRequest('action', {}, 'lpar')
        self.assertEqual('action', preq.action)
        self.assertIsNone(preq.mac_address)
        self.assertEqual('lpar', preq.lpar_uuid)
        self.mock_time.assert_called_once_with()
        self.assertEqual(self.mock_time.return_value, preq.created_at)

    def test_eq(self):
        """Test __eq__ and __ne__."""
        dev1 = {'mac_address': 'mac1', 'some': 'garbage'}
        dev1_2 = {'mac_address': 'mac1', 'other': 'stuff'}
        dev2 = {'mac_address': 'mac2', 'some': 'garbage'}
        # ProvisionRequests with the same mac/LPAR are equal, regardless of
        # other attributes.
        self.assertEqual(
            prov_req.ProvisionRequest('plug', dev1, 'lpar'),
            prov_req.ProvisionRequest('unplug', dev1_2, 'lpar'))
        # ProvisionRequests are not equal if LPAR UUIDs don't match
        self.assertNotEqual(
            prov_req.ProvisionRequest('plug', dev1, 'lpar1'),
            prov_req.ProvisionRequest('plug', dev1, 'lpar2'))
        # ...or if MACs don't match
        self.assertNotEqual(
            prov_req.ProvisionRequest('plug', dev1, 'lpar'),
            prov_req.ProvisionRequest('plug', dev2, 'lpar'))

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'device_detail_valid')
    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.norm_mac')
    def test_for_wrappers(self, mock_mac, mock_ddv):
        def device_detail_valid(detail, mac):
            self.assertEqual(detail['mac_address'], mac)
            # Since we can't rely on the order of the outer loop, fail a
            # predictable mac.
            return mac != 'mac3'
        mock_ddv.side_effect = device_detail_valid
        mock_mac.side_effect = lambda mac: mac
        agent = mock.Mock()
        lpar1, lpar2, lpar3 = mock.Mock(), mock.Mock(), mock.Mock()
        lpar1.configure_mock(name='lpar1', uuid='uuid1')
        lpar2.configure_mock(name='lpar2', uuid='uuid2')
        lpar3.configure_mock(name='lpar3', uuid='uuid3')
        # System returns vifs with macs 1-4
        lpar_vif_map = {
            lpar1: [mock.Mock(mac='mac1'), mock.Mock(mac='mac2')],
            lpar2: [mock.Mock(mac='mac3')],
            lpar3: [mock.Mock(mac='mac4')]}
        # neutron returns devices with macs 1, 3-5.
        # I.e. compared to system, mac2 is absent, and mac5 is extraneous.
        dd1, dd3, dd4, dd5 = [{'mac_address': 'mac%d' % idx} for idx in
                              (1, 3, 4, 5)]
        agent.get_devices_details_list.return_value = [dd1, dd3, dd4, dd5]
        # Run it
        ret = prov_req.ProvisionRequest.for_wrappers(agent, lpar_vif_map, 'ac')
        # We validated 1, 3, and 4.  mac2 was not validated because it wasn't
        # in the neutron device list.  mac5 was not validated because it wasn't
        # in the system vifs.
        mock_ddv.assert_has_calls([mock.call(det, det['mac_address'])
                                   for det in (dd1, dd3, dd4)], any_order=True)
        # We created ProvisionRequests for 1 and 4 (mac3 didn't validate)
        # Sort for easier validation
        ret.sort(key=lambda preq: preq.mac_address)
        self.assertEqual('mac1', ret[0].mac_address)
        self.assertEqual('mac4', ret[1].mac_address)
        self.assertEqual('uuid1', ret[0].lpar_uuid)
        self.assertEqual('uuid3', ret[1].lpar_uuid)
        self.assertEqual('ac', ret[0].action)
        self.assertEqual('ac', ret[1].action)

    @mock.patch('networking_powervm.plugins.ibm.agent.powervm.utils.'
                'device_detail_valid')
    @mock.patch('pypowervm.util.get_req_path_uuid')
    def test_for_event(self, mock_uuid, mock_ddv):
        def mk_evt(action, mac, vif_type, provider='NOVA_PVM_VIF',
                   etype='CUSTOM_CLIENT_EVENT'):
            return mock.Mock(
                detail='{"action": "%s", "mac": "%s", "provider": "%s",'
                       '"type": "%s"}' % (
                           action, mac, provider, vif_type), etype=etype)
        agent = mock.Mock()
        agent.vif_type = 'pvm_sea'

        # Wrong event type
        self.assertIsNone(prov_req.ProvisionRequest.for_event(
            agent, mock.Mock(etype='bogus')))
        # Use this to prove we didn't get further
        agent.get_device_details.assert_not_called()

        # Bogus JSON
        self.assertIsNone(prov_req.ProvisionRequest.for_event(
            agent, mock.Mock(etype='CUSTOM_CLIENT_EVENT', detail=None)))
        agent.get_device_details.assert_not_called()
        self.assertIsNone(prov_req.ProvisionRequest.for_event(
            agent, mock.Mock(etype='CUSTOM_CLIENT_EVENT', detail='')))
        agent.get_device_details.assert_not_called()

        # Wrong provider
        self.assertIsNone(prov_req.ProvisionRequest.for_event(
            agent, mk_evt('action', 'mac', 'pvm_sea', provider='bogus')))
        agent.get_device_details.assert_not_called()

        # Wrong vif_type
        self.assertIsNone(prov_req.ProvisionRequest.for_event(
            agent, mk_evt('action', 'mac', 'pvm_sriov')))
        agent.get_device_details.assert_not_called()

        # Unrecognized action
        self.assertIsNone(prov_req.ProvisionRequest.for_event(
            agent, mk_evt('action', 'mac', 'pvm_sea')))
        agent.get_device_details.assert_not_called()

        # Validation failure
        mock_ddv.return_value = False
        self.assertIsNone(prov_req.ProvisionRequest.for_event(
            agent, mk_evt('plug', 'amac', 'pvm_sea')))
        agent.get_device_details.assert_called_once_with('amac')
        mock_ddv.assert_called_once_with(agent.get_device_details.return_value,
                                         'amac')
        # Prove we didn't get further
        mock_uuid.assert_not_called()

        # Good path
        agent.get_device_details.reset_mock()
        mock_ddv.reset_mock()
        mock_ddv.return_value = True
        evt = mk_evt('unplug', 'amac', 'pvm_sea')
        ret = prov_req.ProvisionRequest.for_event(agent, evt)
        agent.get_device_details.assert_called_once_with('amac')
        mock_ddv.assert_called_once_with(agent.get_device_details.return_value,
                                         'amac')
        mock_uuid.assert_called_once_with(evt.data, preserve_case=True,
                                          root=True)
        self.assertEqual('unplug', ret.action)
        self.assertEqual(agent.get_device_details.return_value, ret.rpc_device)
        self.assertEqual(mock_uuid.return_value, ret.lpar_uuid)
